
/*
 *  Complex - Implements a complex number
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef COMPLEX_H_
#define COMPLEX_H_

#include "vartypes.h"
#include "integer.h"
#include "real.h"
#include "utf8_string.h"


#define COMPLEX_H_VERSION "v3"
#define COMPLEX_H_DATE    "2005-09 - 2006-07"



class Complex
{
public:
	Complex() { real = 0.0; imaginary = 0.0; }
	Complex(const Real &re) throw(error_obj) { real = re; imaginary = Real(0); }
	Complex(const Real &re,const Real &im) throw(error_obj) { real = re; imaginary = im; }
	Complex(const Complex &src) throw(error_obj) { real = src.real; imaginary = src.imaginary; }

	Real real,imaginary;

	inline Complex operator+(const Real &val) const { Complex tmp(real, imaginary); tmp.real.add( val ); return tmp; }
	inline Complex operator+(const Complex &val) const { Complex tmp(real, imaginary); tmp.real.add( val.real ); tmp.imaginary.add( val.imaginary ); return tmp; }
	inline Complex & operator+=(const Real &val) { real.add( val ); return *this; }
	inline Complex & operator+=(const Complex &val) { real.add( val.real ); imaginary.add( val.imaginary ); return *this; }
	inline Complex operator-(const Real &val) const { return Complex(real - val, imaginary); }
	inline Complex operator-(const Complex &var) const { return Complex(real - var.real, imaginary - var.imaginary); }
	inline Complex & operator-=(const Real &val) { real.sub( val ); return *this; }
	inline Complex & operator-=(const Complex &var) { real.sub( var.real ); imaginary.sub( var.imaginary ); return *this; }
	inline Complex operator*(const Real &val) const { return Complex(real * val, imaginary * val); }
	Complex operator*(const Complex &var) const;
	inline Complex operator/(const Real &val) const { return Complex(real / val, imaginary / val); }
	Complex operator/(const Complex &var) const throw(error_obj);

	inline bool operator==(const Complex &var) const { return (real == var.real  &&  imaginary == var.imaginary); }
	inline bool operator!=(const Complex &var) const { return !(*this == var); }
	bool operator<(const Complex &var) const throw(error_obj);
	bool operator>(const Complex &var) const throw(error_obj);
	inline bool operator<=(const Complex &var) const throw(error_obj) { return !(*this > var); }
	inline bool operator>=(const Complex &var) const throw(error_obj) { return !(*this < var); }

	void append_to_string( utf8_string &str, const Format &fmt ) const throw(error_obj);

	inline Complex & operator=(const Complex &src) throw(error_obj) { real = src.real; imaginary = src.imaginary; return *this; }
};
bool complex_almost_equal(const Complex &l,const Complex &r);
Complex power(const Complex &left,const Complex &right) throw(error_obj);
Complex sqroot(const Complex &var) throw(error_obj);
Complex abs(const Complex &var) throw(error_obj);
// arg's return value is in radians
inline Complex arg(const Complex &var) {   return Complex(atan2x(floatx(var.imaginary),floatx(var.real))); }
inline Complex ipart(const Complex &var) { return Complex( ipart(var.real), ipart(var.imaginary) ); }
inline Complex fpart(const Complex &var) { return Complex( fpart(var.real), fpart(var.imaginary) ); }
inline Complex operator+(const Real &val,const Complex &com) { return Complex(com.real + val, com.imaginary); }
inline Complex operator-(const Real &val,const Complex &com) { return Complex(com.real - val, com.imaginary); }
inline Complex operator-(const Complex &com) { return Complex(-com.real, -com.imaginary); }
inline Complex operator*(const Real &val,const Complex &com) { return Complex(com.real * val, com.imaginary * val); }
Complex operator/(const Real &val,const Complex &com);


Complex expc(const Complex &val);
Complex logc(const Complex &val);
Complex log2c(const Complex &val);
Complex log10c(const Complex &val);
Complex sinc(const Complex &val);
Complex cosc(const Complex &val);
Complex tanc(const Complex &val);
Complex asinc(const Complex &val);
Complex acosc(const Complex &val);
Complex atanc(const Complex &val);
Complex sinhc(const Complex &val);
Complex coshc(const Complex &val);
Complex tanhc(const Complex &val);
Complex asinhc(const Complex &val);
Complex acoshc(const Complex &val);
Complex atanhc(const Complex &val);




#endif // COMPLEX_H_
