
/*
 *  Integer - Implements an integer of arbitrary size
 *  Copyright (c) 2005-2007 by Mattias Hultgren <tilda_o_tize@hotmail.com>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef INTEGER_H_
#define INTEGER_H_

#include "vartypes.h"
#include "utf8_string.h"


#define INTEGER_H_VERSION "v7"
#define INTEGER_H_DATE    "2005-04 - 2006-10"


class Integer
{
private:
	uint32 size, used;
	uint32 buf[2];
	uint32 *data;
	bool negative;

public:
	Integer();
	Integer( const Integer &src ) throw(error_obj);
	Integer(int32 src);
	Integer(int64 src);
	Integer(floatx src) throw(error_obj);
	Integer & operator=(floatx val) throw(error_obj);
	operator floatx(void);
	~Integer();

	void enlarge_to( uint32 new_size ) throw(error_obj);

	void extract( Integer *dest, uint32 start, uint32 stop ) const throw(error_obj);

	uint32 get_used(void) const { return used; }
	const uint32* get_data(void) const { return data; }


// *_data are low level functions, and should not be called unless
// you are sure what you're doing
	void add_data( const Integer &nr ) throw(error_obj);
	void add_data_uint32_shifted( const Integer &nr, uint32 shift) throw(error_obj); // this += nr * 2^(32*shift)
	// sub_data:  this must be larger than nr....no checking for that will be done
	void sub_data( const Integer &nr ); // this -= nr
	void sub_data_reverse( const Integer &nr ); // this = nr - this
	void mul_data( const Integer &factor, Integer *dest ) const throw(error_obj);

	void add( const Integer &val ) throw(error_obj);
	void sub( const Integer &val ) throw(error_obj);
	void mul( const Integer &factor, Integer *dest ) const throw(error_obj);

	void mul_by_2(void) throw(error_obj);
	void mul_by_2_pow_32(uint32 n=1) throw(error_obj);
	void div_by_2(void) throw(error_obj);
	void div_by_2_pow_32(void) throw(error_obj);
	void divide( const Integer &divider, Integer *ans, Integer *rest ) const throw(error_obj);

	// special function, used by divide      *this -= nr * 2^(32*shift) * mult
	// this must be larger than   nr * 2^(32*shift) * mult
	void sub_uint32_mult_shifted( uint32 mult, const Integer &nr, uint32 shift );

	// special function, used by divide      *this += nr * 2^(32*shift)
	// this must be large enough and the data should be zero filled
	void add_uint32_shifted( uint32 nr, uint32 shift ) throw(error_obj);

	uint32 count_used_bits(void) const;

	void set_bit( uint32 bit ) throw(error_obj);
	void clear_bit( uint32 bit ) throw(error_obj);

	bool check_bit( uint32 bit ) const throw(error_obj);

	Integer operator+(const Integer &val) const throw(error_obj);
	Integer operator-(const Integer &val) const throw(error_obj);
	inline Integer operator*(const Integer &val) const throw(error_obj) { Integer ans; mul(val, &ans); return ans; }
	inline Integer operator/(const Integer &val) const throw(error_obj) { Integer ans; divide( val, &ans, 0 ); return ans; }
	inline Integer operator%(const Integer &val) const throw(error_obj) { Integer rest; divide( val, 0, &rest ); return rest; }

	inline void swap( Integer &nr )
		{
			bool t;
			uint32 t2, *t3;

			t = negative;
			negative = nr.negative;
			nr.negative = t;

			for( int i=0; i<2; i++ )
			{
				t2 = nr.buf[i];
				nr.buf[i] = buf[i];
				buf[i] = t2;
			}

			t2 = size;
			size = nr.size;
			nr.size = t2;

			t2 = used;
			used = nr.used;
			nr.used = t2;

			t3 = (data==buf) ? nr.buf : data;
			data = (nr.data==nr.buf) ? buf : nr.data;
			nr.data = t3;
		}

	Integer & operator=(const Integer &val) throw(error_obj);
	Integer & operator=(int32 val);
	Integer & operator=(int64 val);
	inline Integer & operator+=(const Integer &val) throw(error_obj) { add( val ); return *this; }
	inline Integer & operator-=(const Integer &val) throw(error_obj) { sub( val ); return *this; }
	inline Integer & operator*=(const Integer &val) throw(error_obj) { *this = *this * val; return *this; }
	inline Integer & operator/=(const Integer &val) throw(error_obj) { *this = *this / val; return *this; }
	inline Integer & operator%=(const Integer &val) throw(error_obj) { *this = *this % val; return *this; }

	bool operator==(const Integer &val) const;
	inline bool operator!=(const Integer &val) const { return !(*this == val); }
	bool operator==(int32 val) const;
	inline bool operator!=(int32 val) const { return !(*this == val); }

	bool operator<(const Integer &val) const;
	bool operator>(const Integer &val) const;
	inline bool operator<=(const Integer &val) const { return !(*this > val); }
	inline bool operator>=(const Integer &val) const { return !(*this < val); }
	bool operator<(int32 val) const;
	bool operator>(int32 val) const;
	inline bool operator<=(int32 val) const { return !(*this > val); }
	inline bool operator>=(int32 val) const { return !(*this < val); }

	// -1 this is smaller
	//  0 they are equal
	//  1 this is bigger
	int compare_abs_value( const Integer &val ) const;

	//  same as compare_abs_value, but val is temporary shifted left shift-steps
	// compare_abs_value_shifted( val, shift )
	//  equals
	// val.mul_by_2_pow_32( shift );
	// compare_abs_value( val );
	// val.div_by_2_pow_32( shift );
	int compare_abs_value_shifted( const Integer &val, uint32 shift ) const;

	Integer & operator++(void) throw(error_obj);
	inline Integer & operator++(int) throw(error_obj) { return ++ *this; }
	Integer & operator--(void) throw(error_obj);
	inline Integer & operator--(int) throw(error_obj) { return -- *this; }

	friend Integer operator-(const Integer &val) throw(error_obj);

	inline void negate(void) { negative = !negative; }
	inline bool isNegative(void) const { return negative; }
	inline bool isZero(void) const { return (used == 0); }

	operator int64(void);

	void append_to_string( utf8_string &str, uint32 nrofdigits=0 ) const throw(error_obj);

	void power( Integer nr, Integer *dest ) const throw(error_obj);

	void faculty(void) throw(error_obj);

	void perm( Integer n, Integer k ) throw(error_obj);
	void comb( const Integer &up, Integer down ) throw(error_obj);


	friend Integer iSqrt( const Integer &var ) throw(error_obj);

	// gcd - greatest common divisior (using Euklides algorithm)
	// if any of nr1 & nr2 is zero the return value is zero also
	friend Integer gcd( Integer nr1, Integer nr2 ) throw(error_obj);

	// lcm - least common multiple ( (nr1*nr2)/gcd(nr1,nr2) )
	// if(gcd(nr1,nr2) == 0) the return value is zero also
	friend Integer lcm( const Integer &nr1, const Integer &nr2 ) throw(error_obj);
};


// dest = [ (n^p) mod m ]
void powMod( Integer *dest, Integer n, Integer p, const Integer &m ) throw(error_obj);


#endif // INTEGER_H_
