
/*
 *  Math2
 *  Copyright (c) 2004-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See math2.h
 */


#include "math2.h"
#include "math2.intern.h"
#include "keyfile.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

namespace math
{

int count_whitespaces(const char *str, int *line_breaks = 0);
void overwrite_comments(char *str) throw(error_obj);
int get_word( const char *str, utf8_string &word ) throw(error_obj);
void overwrite_script_info(char *str);


const char *reserved_keywords[] = {
                                   "Function",
                                   "Variable",
                                   "Integer",
                                   "Real",
                                   "Complex",
                                   "String",
                                   "Matrix",
                                   "Picture",
                                   "Array",
                                   "Boolean",
                                   "return",
                                   "if",
                                   "else",
                                   "while",
                                   "break",
                                   "continue",
                                   0
                                  };


void check_name_against_reserved(const utf8_string &name) throw(error_obj)
{
	for(int i=0;;i++)
	{
		if(reserved_keywords[i] == 0)
			break;

		if( name == reserved_keywords[i] )
		{
			utf8_string tmpstr( _("Illegal use of reserved keyword: ") );
			tmpstr.append( name );
			THROW_ERROR( ErrorType_General, tmpstr.c_str() );
		}
	}
}



CodeBlock::CodeBlockLine::CodeBlockLine()
{
	line_number = 0;
	type = CodeBlock_Type_Start;
	sub_type = CodeBlock_SubType_None;
	code = 0;
}
CodeBlock::CodeBlockLine::CodeBlockLine(const CodeBlockLine &src) throw(error_obj)
{
	line_number = 0;
	type = CodeBlock_Type_Start;
	sub_type = CodeBlock_SubType_None;
	code = 0;

	*this = src;
}
CodeBlock::CodeBlockLine::~CodeBlockLine()
{
	if( type == CodeBlock_Type_Code )
		delete code;
}
void CodeBlock::CodeBlockLine::operator=(const CodeBlockLine &src) throw(error_obj)
{
	if( src.code != 0  &&  src.type == CodeBlock_Type_Code )
	{
		if(code == 0)
		{
			try{  code = new CodeLine;  }
			catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
		}

		try{  *code = *src.code;  }
		catch(...)
		{
			delete code;
			code = 0;
			throw;
		}
	}
	else
	{
		if( type == CodeBlock_Type_Code )
		{
			delete code;
			code = 0;
		}
	}

	line_number = src.line_number;
	type = src.type;
	sub_type = src.sub_type;
}






CodeBlock::CodeBlock(const CodeBlock &src) throw(error_obj)
{
	inuse = 0;
	nroftmpvariables = 0;
	*this = src;
}

CodeBlock::~CodeBlock()
{
	lines.clear();
}

void CodeBlock::set_code(const char *newcode) throw(error_obj)
{
	char *str;

	clear();

	try{  str = new char [strlen(newcode) + 1];  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
	strcpy(str,newcode);

	try{ set_code_intern( str, true ); }
	catch(...)
	{
		delete [] str;
		throw;
	}
	delete [] str;
}

void CodeBlock::set_code_from_file( const utf8_string &new_filename,
                                    const utf8_string *only_allow_identifier ) throw(error_obj)
{
	FILE *fp;
	long filesize;
	char *str;

	clear();

	filename = new_filename;

	fp = fopen( filename.c_str(), "rb" );
	if(fp == 0)
	{
setcode_thrower_error:
		if(fp != 0)
			fclose(fp);

		clear();
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/read from file (%s)"), new_filename.c_str() );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	if(fseek(fp,0L,SEEK_END) != 0)
		goto setcode_thrower_error;
	filesize = ftell(fp);
	if(filesize == -1)
		goto setcode_thrower_error;
	if(fseek(fp,0L,SEEK_SET) != 0)
		goto setcode_thrower_error;

	try{  str = new char [filesize + 1];  }
	catch(...)
	{
		fclose(fp);
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}

	if(long(fread(str,1,filesize,fp)) != filesize)
	{
		delete [] str;
		goto setcode_thrower_error;
	}
	if(fclose(fp) != 0)
	{
		delete [] str;
		clear();
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't close file (%s)"), new_filename.c_str() );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	str[filesize] = 0;
	try
	{
		overwrite_comments( str );

		while( create_function( str, only_allow_identifier ) ) { }

		if( str[ count_whitespaces(str) ] != '\0' )
		{
			char tmp_err[ERROR_OBJ_MSG_LEN];
			snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Invalid format in file %s"), filename.c_str() );
			THROW_ERROR( ErrorType_File_IO, tmp_err );
		}
	}
	catch(...)
	{
		delete [] str;
		clear();
		throw;
	}

	delete [] str;
	clear();
}

void CodeBlock::calc( Variable *answer ) throw(error_obj)
{
	int i;
	CodeBlock_SubType tmp = CodeBlock_SubType_None;
	Vector<Variable> tmp_variables;
	VariableList this_private_varlist;

	tmp_variables.set_size_filled( nroftmpvariables );

	if( answer == 0 )
		THROW_ERROR( ErrorType_Internal, "No place to store answer in CodeBlock::calc()" );

	try
	{
		inuse++;

		i = 0;
		if( private_varlist.is_empty() )
			run_block( &tmp_variables, 0, tmp, answer, i );
		else
		{
			this_private_varlist = private_varlist; // makes a private copy of the local variable for the function

			run_block( &tmp_variables, &this_private_varlist, tmp, answer, i );
		}

		switch( tmp )
		{
		case CodeBlock_SubType_None:
		case CodeBlock_SubType_Return:
			break;
		
		case CodeBlock_SubType_Break:
			THROW_ERROR( ErrorType_General, _("break used outside of while") );
		
		case CodeBlock_SubType_Continue:
			THROW_ERROR( ErrorType_General, _("continue used outside of while") );
		
		default:
			{
				char str[50];
				snprintf( str, 50, "Unhandled CodeBlock_SubType return %ld", tmp ); 
			
				THROW_ERROR( ErrorType_Internal, str );
			}
		}
	}
	catch(error_obj error)
	{
		inuse--;
		if( filename.get_length() == 0 )
			throw error;
		
		if( error.type >= ErrorType_UmpCode_General )
			throw error;

		error_obj error2;
		error2.type = error.type;

		snprintf( error2.msg, ERROR_OBJ_MSG_LEN, "%s: %s %ld: %s", filename.c_str(), _("line"),
		          lines[i]->line_number, error.msg );
		throw error2;
	}
	catch(...) {  inuse--;  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	inuse--;
}

void CodeBlock::clear(void)
{
	try{
		filename = "";
		lines.clear();
		nroftmpvariables = 0;
	}
	catch(...) { }
}

void CodeBlock::operator=(const CodeBlock &src) throw(error_obj)
{
	printf("\nCodeBlock::operator=() not implemented yet\n");
	exit(-1);
}


int count_whitespaces(const char *str, int *line_breaks)
{
	for(int i=0;;i++)
	{
		switch(str[i])
		{
		case '\n':
			if( line_breaks != 0 )
				line_breaks[0]++;
		case ' ': case '\t':
			break;
		default:
			return i;
		}
	}
}

// this function overwrite all comments with spaces ( \n inside comments are not moved)
// \t is also overwritten by one space
void overwrite_comments(char *str) throw(error_obj)
{
	int i;
	bool inside_string = false;

	for(i=0;str[i] != '\0';i++)
	{
		if(inside_string)
		{
			if( str[i] == '\"'  ||  str[i] == '\n' )
				inside_string = false;
		}
		else
		{
			switch(str[i])
			{
			case '\"':
				inside_string = true;
				break;

			case '\t':
				str[i] = ' ';
				break;

			case '#':
				for(;;i++) // inside a one line comment
				{
					str[i] = ' ';
					if( str[i+1] == '\0'  ||  str[i+1] == '\n' )
						break;
				}
				break;
			
			case '/':
				if(str[i+1] == '/') // inside a one line comment
				{
					for(;;i++)
					{
						str[i] = ' ';
						if( str[i+1] == '\0'  ||  str[i+1] == '\n' )
							break;
					}
				}
				else if(str[i+1] == '*') // inside multi line comment
				{
					for(;;i++)
					{
						if(str[i] != '\n')
							str[i] = ' ';
						if(str[i+1] == '\0')
							THROW_ERROR( ErrorType_General, _("Unexpected end of file") );
						if( str[i+1] == '*'  &&  str[i+2] == '/' )
						{
							str[i+1] = ' ';
							str[i+2] = ' ';
							break;
						}
					}
				}
				break;

			default:
				break;
			}
		}
	}
}

int get_word( const char *str, utf8_string &word ) throw(error_obj)
{
	char tknstr[] = " ";
	int i=0;

	word = "";
	if( str[i] >= '0'  &&  str[i] <= '9' )
		return 0;
	for(;;i++)
	{
		if( (str[i] >= 'A'  &&  str[i] <= 'Z')  ||  (str[i] >= 'a'  &&  str[i] <= 'z')  ||  str[i] == '_'  ||  (str[i] >= '0'  &&  str[i] <= '9') )
		{
			tknstr[0] = str[i];
			word.append( tknstr );
		}
		else
			break;
	}

	return i;
}

void CodeBlock::set_code_intern(char *str, bool allow_variable_creation ) throw(error_obj)
{
	utf8_string word;
	int i, i_tmp;
	int start = 0;
	int linenr = 1;
	int started_curley = 0;
	int started_bracket = 0;
	bool inside_string = false;
	bool end_by_bracket = false;
	CodeBlock_SubType sub_type = CodeBlock_SubType_None;

	try
	{
		for(i=0;;i++)
		{
			if( i == start )
			{
				i += count_whitespaces( &str[i], &linenr );
				i_tmp = i + get_word( &str[i], word );
				if( word == reserved_keywords[reserved_break] )
				{
					i = i_tmp;
					start = i;
					
					i += count_whitespaces( &str[i], &linenr );
					if( str[i] == ';'  ||  str[i] == '{'  ||  str[i] == '}' )
						add_void( CodeBlock_SubType_Break );
					else
					{
						utf8_string tmpstr( _("Illegal use of reserved keyword: ") );
						tmpstr.append( "break" );
						THROW_ERROR( ErrorType_General, tmpstr.c_str() );
					}
				}
				else if( word == reserved_keywords[reserved_continue] )
				{
					i = i_tmp;
					start = i;
					
					i += count_whitespaces( &str[i], &linenr );
					if( str[i] == ';'  ||  str[i] == '{'  ||  str[i] == '}' )
						add_void( CodeBlock_SubType_Continue );
					else
					{
						utf8_string tmpstr( _("Illegal use of reserved keyword: ") );
						tmpstr.append( "continue" );
						THROW_ERROR( ErrorType_General, tmpstr.c_str() );
					}
				}
				else if( word == reserved_keywords[reserved_return] )
				{
					i = i_tmp;
					start = i;
					
					i += count_whitespaces( &str[i], &linenr );
					if( str[i] == ';'  ||  str[i] == '{'  ||  str[i] == '}' )
						add_void( CodeBlock_SubType_Return );
					else
						sub_type = CodeBlock_SubType_Return;
				}
				else if( word == reserved_keywords[reserved_while] )
				{
					i = i_tmp;
					i += count_whitespaces( &str[i], &linenr );

					if( str[i] != '(' )
						THROW_ERROR( ErrorType_General, _("Expected ( after while") );

					started_bracket = 1;
					end_by_bracket = true;
					sub_type = CodeBlock_SubType_While;

					i++;
					start = i;
				}
				else if( word == reserved_keywords[reserved_if] )
				{
					i = i_tmp;
					i += count_whitespaces( &str[i], &linenr );

					if( str[i] != '(' )
						THROW_ERROR( ErrorType_General, _("Expected ( after if") );

					started_bracket = 1;
					end_by_bracket = true;
					sub_type = CodeBlock_SubType_If;

					i++;
					start = i;
				}
				else if( word == reserved_keywords[reserved_else] )
				{
					i = i_tmp;
					i += count_whitespaces( &str[i], &linenr );

					add_void( CodeBlock_SubType_Else );

					if( str[i] == '{' )
					{
						add_curley( CodeBlock_Type_Start );
						started_curley++;
						i++;
					}
					else if( str[i] == ';' )
					{
						add_curley( CodeBlock_Type_Start ); // adding an empty body
						add_curley( CodeBlock_Type_Stop );
						i++;
					}
					start = i;
					i--;
					continue;
				}
			}

			if(inside_string)
			{
				if(str[i] == '\"')
					inside_string = false;
				else if( str[i] == '\n'  ||  str[i] == '\0' )
					THROW_ERROR( ErrorType_General, _("Expected ending \"") );
			}
			else
			{
				switch(str[i])
				{
				case '\n':
					linenr++;
					break;

				case '\"':
					inside_string = true;
					break;

				case '(':
					started_bracket++;
					break;
				
				case ')':
					started_bracket--;
					if( started_bracket == 0  &&  end_by_bracket )
					{
						end_by_bracket = false;
						str[i] = '\0';
						add_line( &str[start], linenr, allow_variable_creation, sub_type );
						sub_type = CodeBlock_SubType_None;
						str[i] = ')';
						i += 1 + count_whitespaces( &str[i+1], &linenr );
						
						if( str[i] == ';' )
						{
							add_curley( CodeBlock_Type_Start ); // adding an empty body
							add_curley( CodeBlock_Type_Stop );
							start = i+1;
						}
						else
						{
							start = i;
							i--;
						}
					}
					break;
				
				case '{':
					if( end_by_bracket )
						THROW_ERROR( ErrorType_General, _("Expected ) but found {") );
					if(start != i)
					{
						str[i] = '\0';
						add_line( &str[start], linenr, allow_variable_creation, sub_type );
						sub_type = CodeBlock_SubType_None;
						str[i] = '{';
					}
					add_curley( CodeBlock_Type_Start );

					started_curley++;

					start = i+1;
					break;

				case '}':
					if( end_by_bracket )
						THROW_ERROR( ErrorType_General, _("Expected ) but found }") );
					if(start != i)
					{
						str[i] = '\0';
						add_line( &str[start], linenr, allow_variable_creation, sub_type );
						sub_type = CodeBlock_SubType_None;
						str[i] = '}';
					}
					add_curley( CodeBlock_Type_Stop );

					if(started_curley == 0)
					{
						clear();
						THROW_ERROR( ErrorType_General, _("Unexpected }") );
					}
					started_curley--;

					start = i+1;
					break;

				case ';':
					if( end_by_bracket )
						THROW_ERROR( ErrorType_General, _("Expected ) but found ;") );
					if(start != i)
					{
						str[i] = '\0';
						add_line( &str[start], linenr, allow_variable_creation, sub_type );
						sub_type = CodeBlock_SubType_None;
						str[i] = ';';
					}
					start = i+1;
					break;

				case '\0':
					if( end_by_bracket )
						THROW_ERROR( ErrorType_General, _("Expected )") );
					if(start != i)
						add_line( &str[start], linenr, allow_variable_creation, sub_type );
					break;
				}
			}
	
			if(str[i] == '\0')
				break;
		}
		if(started_curley != 0)
			THROW_ERROR( ErrorType_General, _("Expected }") );


		add_automatic_curleys();


		for( i=0; i<lines.get_size(); i++ ) // let's check for malplaced else's
		{
			if( lines[i]->sub_type == CodeBlock_SubType_Else )
			{
				if( i == 0 )
					THROW_ERROR( ErrorType_General, _("Unexpected 'else'") );
				
				if( lines[i-1]->type == CodeBlock_Type_Stop )
				{
					if( lines[i-1]->pair_with == 0 )
						THROW_ERROR( ErrorType_General, _("Unexpected 'else'") );
					if( lines[ lines[i-1]->pair_with -1 ]->sub_type != CodeBlock_SubType_If )
						THROW_ERROR( ErrorType_General, _("Unexpected 'else'") );
				}
				else
					THROW_ERROR( ErrorType_General, _("Unexpected 'else'") );
			}
		}

		for( i=0, nroftmpvariables=0; i<lines.get_size(); i++ ) // updating nroftmpvariables
		{
			if( lines[i]->type == CodeBlock_Type_Code )
				nroftmpvariables = (nroftmpvariables > lines[i]->code->get_nroftmpvariables()) ? nroftmpvariables : lines[i]->code->get_nroftmpvariables();
		}
	}
	catch(error_obj error)
	{
		if( filename.get_length() == 0 )
		{
			clear();
			throw error;
		}

		error_obj error2;
		error2.type = error.type;

		snprintf( error2.msg, ERROR_OBJ_MSG_LEN, "%s: %s %ld: %s", filename.c_str(), _("line"),
		          linenr, error.msg );

		clear();
		throw error2;
	}
}

void CodeBlock::add_automatic_curleys(void) throw(error_obj)
{
	bool added_curley;
	int i, start;

	i = lines.get_size()-1;
	if( i == -1 )
		return;

	if( lines[i]->sub_type == CodeBlock_SubType_While  ||
	    lines[i]->sub_type == CodeBlock_SubType_If  ||
	    lines[i]->sub_type == CodeBlock_SubType_Else )
		THROW_ERROR( ErrorType_General, _("Expected {") );

	do
	{
		added_curley = false;
		pair_up_curleys();
		
		for( i=lines.get_size()-2; i>=0; i-- )
		{
			if( lines[i+1]->type != CodeBlock_Type_Start )
			{
				if( lines[i]->sub_type == CodeBlock_SubType_While  ||
				    lines[i]->sub_type == CodeBlock_SubType_If  ||
				    lines[i]->sub_type == CodeBlock_SubType_Else )
				{
					i++;
					start = i;

					if( lines[i]->sub_type == CodeBlock_SubType_While )
						i = lines[i+1]->pair_with +1;
					else if( lines[i]->sub_type == CodeBlock_SubType_If )
					{
						i = lines[i+1]->pair_with +1;
						if( lines[i]->sub_type == CodeBlock_SubType_Else )
							i = lines[i+1]->pair_with +1;
					}
					else
						i++;

					add_curley( CodeBlock_Type_Stop, CodeBlock_SubType_None, i );
					add_curley( CodeBlock_Type_Start, CodeBlock_SubType_None, start );
					added_curley = true;
					break;
				}
			}
		}
	}while( added_curley );
}

void CodeBlock::pair_up_curleys(void)
{
	int i, i_tmp;
	for( i=0; i<lines.get_size(); i++ )
	{
		if( lines[i]->type == CodeBlock_Type_Start  ||
		    lines[i]->type == CodeBlock_Type_Stop )
			lines[i]->pair_with = -1;
	}
	for( i=0; i<lines.get_size(); i++ )
	{
		if( lines[i]->type == CodeBlock_Type_Stop )
		{
			if( lines[i]->pair_with == -1 )
			{
				i_tmp = i;
				for( i--; i>=0; i-- )
				{
					if( lines[i]->type == CodeBlock_Type_Start )
					{
						if( lines[i]->pair_with == -1 )
						{
							lines[i]->pair_with = i_tmp;
							lines[i_tmp]->pair_with = i;
							break;
						}
					}
				}
			}
		}
	}
}

// return true if a function was created (and overwrited with spaces, \n are left as is)
// if an identifier was found that didn't match with only_allow_identifier, false is returned
bool CodeBlock::create_function(char *str, const utf8_string *only_allow_identifier) throw(error_obj)
{
	int i,i_tmp;
	utf8_string name, word;
	CodeBlock *func;
	Vector<ArgumentType> arguments;
	ArgumentType tmp_argument;

	try
	{
		func = new CodeBlock;
	}
	catch(...)  {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	try
	{
		i = count_whitespaces( str ); // skipping whitespaces
		if(str[i] == '\0')
		{
			delete func;
			return false;
		}

// Checking for "Function"
		i += get_word( &str[i], word );
		if( word != reserved_keywords[reserved_function] )
		{
			delete func;
			return false;
		}

		if( (i_tmp = count_whitespaces( &str[i] )) ==  0 )
		{
			delete func;
			return false; // keyword "Function" must be followed by whitespace(s)
		}
		i += i_tmp;
// Found "Function"

// Getting name of function
		i += get_word( &str[i], name );

// Check so that the name is allowed
		if( only_allow_identifier != 0 )
		{
			if( name != *only_allow_identifier )
			{
				delete func;
				return false;
			}
		}
		
// Make sure the name is legal and unique
		check_name_against_reserved( name );

		if( function_list.check_for( name ) )
		{
			error_obj error;

			error.type = ErrorType_Name_isnt_unique;
			snprintf( error.msg, ERROR_OBJ_MSG_LEN, "Function name (%s) is not unique", name.c_str() );
			throw error;
		}

		try
		{
			global_varlist.get_id( name.c_str() ); // this line will throw if the name isn't a Variable
			error_obj error;
			error.type = ErrorType_Name_isnt_unique;
			snprintf( error.msg, ERROR_OBJ_MSG_LEN, "Function name (%s) is not unique", name.c_str() );
			throw error;
		}
		catch(error_obj error) // only throwed error is 'Variable not found' and 'Function name () is not unique'
		{
			if( error.type == ErrorType_Name_isnt_unique )
				throw;
		}
// The name is now legal and unique
		
// Now its time for the argument brackets
		i += count_whitespaces( &str[i] );
		if(str[i] == '(')  // checking if the function should have arguments
		{
			for(i++;;)
			{
				i += count_whitespaces( &str[i] );

				get_word( &str[i], word );
				if( word.get_length() != 0 )
				{
					if( word == reserved_keywords[reserved_variable] )
					{
						tmp_argument.type = reserved_variable;
specified_argument_entry:
						i += word.get_length();
						i += count_whitespaces( &str[i] );
						get_word( &str[i], word );
unspecified_argument_entry:
						i += word.get_length();
						i += count_whitespaces( &str[i] );

						check_name_against_reserved( word );
						if( word == "i" )
							THROW_ERROR( ErrorType_General, _("\'i\' isn't a valid name of an argument variable.") );

						try{
							func->private_varlist.get_id( word.c_str() );
							word.prepend( _("Name is not unique: ") );
							THROW_ERROR( ErrorType_Name_isnt_unique, word.c_str() );
						}
						catch(error_obj error) // only throwed error is 'Variable not found' and 'Name is not unique'
						{
							if( error.type == ErrorType_Name_isnt_unique )
								throw;

							try{  tmp_argument.variable_name = word;  }
							catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
							
							arguments.append( tmp_argument );

							Variable tmpvar;
							tmpvar.set_name( word );
							func->private_varlist.create( tmpvar, false, false, true );
						}
						if( str[i] == ')' )
							break;
						else if( str[i] == ',' )
							i++;
					}
					else if( word == reserved_keywords[reserved_integer] )
					{
						tmp_argument.type = reserved_integer;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_real] )
					{
						tmp_argument.type = reserved_real;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_complex] )
					{
						tmp_argument.type = reserved_complex;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_string] )
					{
						tmp_argument.type = reserved_string;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_matrix] )
					{
						tmp_argument.type = reserved_matrix;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_picture] )
					{
						tmp_argument.type = reserved_picture;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_array] )
					{
						tmp_argument.type = reserved_array;
						goto specified_argument_entry;
					}
					else if( word == reserved_keywords[reserved_boolean] )
					{
						tmp_argument.type = reserved_boolean;
						goto specified_argument_entry;
					}
					else
					{
// this is an unspecified argument make it the same as Variable
						tmp_argument.type = reserved_variable;
						goto unspecified_argument_entry;
					}
				}
				else
				{
					if( arguments.get_size() != 0 )
						THROW_ERROR( ErrorType_General, _("Expected argument") );

					if( str[i] == '\0' )
					{
						THROW_ERROR( ErrorType_General, _("Expected )") );
					}
					else if( str[i] == ')' )
						break;
					else
						THROW_ERROR( ErrorType_General, _("Syntax error") );
				}
			}
			i += 1 + count_whitespaces( &str[i+1] );
		}
// Argument brackets handled

// Function body start
		if(str[i] != '{')
			THROW_ERROR( ErrorType_General, _("Expected {") );
		i++;

// The functions intern variables should be created
		for(;;)
		{
			i += count_whitespaces( &str[i] );

			get_word( &str[i], word );
			i += word.get_length();
			if( word == reserved_keywords[reserved_variable] )
			{
				do
				{
					i += count_whitespaces( &str[i] );

					i += get_word( &str[i], word );

					check_name_against_reserved( word );
					if( word == "i" )
						THROW_ERROR( ErrorType_General, _("\'i\' isn't a valid name of an intern variable.") );

					try{
						func->private_varlist.get_id( word.c_str() );

						word.prepend( _("Name is not unique: ") );
						THROW_ERROR( ErrorType_Name_isnt_unique, word.c_str() );
					}
					catch(error_obj error) // only throwed error is 'Variable not found' and 'Name is not unique'
					{
						if( error.type == ErrorType_Name_isnt_unique )
							throw;

						Variable tmpvar;
						tmpvar.set_name( word );
						func->private_varlist.create( tmpvar, false, false, true );
					}
					i += count_whitespaces( &str[i] );
					if( str[i] == ',' )
						i++;
					else if( str[i] == ';' )
					{
						i++;
						break;
					}
					else
						THROW_ERROR( ErrorType_General, _("Expected ;") );
				}while( true );
			}
			else
			{
				i -= word.get_length();
				break; 
			}
		}
// All intern variable has been created


		{
			bool inside_string = false;
			int nested_blocks = 0;
			for(i_tmp=0;;i_tmp++) // getting the length of the function
			{
				if(str[i+i_tmp] == '\0')
					THROW_ERROR( ErrorType_General, _("Expected }") );

				if(inside_string)
				{
					if(str[i+i_tmp] == '\"')
						inside_string = false;
				}
				else
				{
					if(str[i+i_tmp] == '{')
						nested_blocks++;
					else if(str[i+i_tmp] == '}')
					{
						if(nested_blocks == 0)
							break;
						else
							nested_blocks--;
					}
					else if(str[i+i_tmp] == '\"')
						inside_string = true;
				}
			}
		}
// Function body ended

		for(int i2=0; i2 < i; i2++ ) // overwriting  "Function name(){" with spaces
		{
			if( str[i2] != '\n' )
				str[i2] = ' ';
		}

	}
	catch(error_obj error)
	{
		error_obj error2;
		int linenr = 1;
		error2.type = error.type;

		delete func;

		for( i_tmp=0; i_tmp < i; i_tmp++ )
		{
			if( str[i_tmp] == '\n' )
				linenr++;
		}

		snprintf( error2.msg, ERROR_OBJ_MSG_LEN, "%s: %s %ld: %s", filename.c_str(), _("line"),
		          linenr, error.msg );
		throw error2;
	}


	str[i+i_tmp] = '\0'; // overwriting the ending '}'

	try{  func->filename = filename;  }
	catch(...) { } // don't care about this error, 'cause it just disables filename in error-msg

	try
	{
// Temporary adding a dummy function (to be able to have recursive calls)
		function_list.add_this_code_block_function( name, 0, arguments );

		try{  func->set_code_intern( str, false );  }
		catch(...)
		{
			function_list.delete_function( name );
			throw;
		}
		
// Removing the dummy function added
		function_list.delete_function( name );

		function_list.add_this_code_block_function( name, func, arguments );
	}
	catch(...)
	{
		delete func;
		throw;
	}

	for(int i2=i; i2 <= (i+i_tmp); i2++ ) // overwriting the whole function with spaces
	{
		if( str[i2] != '\n' )
			str[i2] = ' ';
	}

	return true;
}

void CodeBlock::add_line(const char *codeline, int linenr, bool allow_variable_creation,
                         CodeBlock_SubType sub_type) throw(error_obj)
{
	CodeBlockLine *line = 0;

	try
	{
		try
		{
			line = new CodeBlockLine;
			lines.append_this( line );
		}
		catch(...)
		{
			delete line;
			throw;
		}

		line->line_number = linenr;
		line->type = CodeBlock_Type_Code;
		line->sub_type = sub_type;

		line->code = new CodeLine;
		if( private_varlist.is_empty() )
			line->code->set_code_line( codeline, allow_variable_creation, 0 );
		else
			line->code->set_code_line( codeline, allow_variable_creation, &private_varlist );

		if( line->code->is_empty() )
			lines.remove( lines.get_size()-1 );
	}
	catch(error_obj error) {  throw;  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}

void CodeBlock::add_curley(CodeBlock_Type type, CodeBlock_SubType sub_type, int pos) throw(error_obj)
{
	CodeBlockLine *line = 0;

	try
	{
		try
		{
			line = new CodeBlockLine;
			lines.insert_this( line, pos );
		}
		catch(...)
		{
			delete line;
			throw;
		}

		line->type = type;
		line->sub_type = sub_type;
		line->pair_with = -1;
	}
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}

void CodeBlock::add_void( CodeBlock_SubType sub_type ) throw(error_obj)
{
	CodeBlockLine *line = 0;

	try
	{
		try
		{
			line = new CodeBlockLine;
			lines.append_this( line );
		}
		catch(...)
		{
			delete line;
			throw;
		}

		line->type = CodeBlock_Type_Void;
		line->sub_type = sub_type;
	}
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}

void CodeBlock::run_block( Vector<Variable> *tmp_variables, VariableList *this_private_varlist, CodeBlock_SubType &ret,
                           Variable *ans, int &start, bool expect_curley_ending ) const throw(error_obj)
{
	Variable *tmp;

	for( ; start<lines.get_size(); start++ )
	{
		switch( lines[start]->type )
		{
		case CodeBlock_Type_Code:
			if( start == (lines.get_size()-1) )
				tmp = ans;
			else
			{
				switch( lines[start]->sub_type )
				{
				case CodeBlock_SubType_If:
				case CodeBlock_SubType_While:
				case CodeBlock_SubType_Return:
					tmp = ans;
					break;
				default:
					tmp = 0;
				}
			}
			lines[start]->code->calc( tmp, this_private_varlist, tmp_variables );

			if( tmp != 0 )
			{
				switch( lines[start]->sub_type )
				{
				case CodeBlock_SubType_Return:
					ret = CodeBlock_SubType_Return;
					return;

				case CodeBlock_SubType_While:
					{
						int tmp;

						if( ans->get_type() != VariableType_Boolean )
						{
							char tmp_err[ERROR_OBJ_MSG_LEN];
							snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Data type is unsupported by '%s'"), "while" );
							THROW_ERROR( ErrorType_General, tmp_err );
						}
					
						if( ans->get_boolean() )
						{  // lets run the 'while'-block
							tmp = start;
							start += 2;
							run_block( tmp_variables, this_private_varlist, ret, ans, start, true );
							switch( ret )
							{
							case CodeBlock_SubType_None:
							case CodeBlock_SubType_Continue:
								start = tmp-1;
								break;

							case CodeBlock_SubType_Break:
								start = lines[tmp+1]->pair_with;
								break;

							case CodeBlock_SubType_Return:
								return;
							}
							ret = CodeBlock_SubType_None;
						}
						else  // lets skip the 'while'-block
							start = lines[start+1]->pair_with;
					}
					break;

				case CodeBlock_SubType_If:
					if( ans->get_type() != VariableType_Boolean )
					{
						char tmp_err[ERROR_OBJ_MSG_LEN];
						snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Data type is unsupported by '%s'"), "if" );
						THROW_ERROR( ErrorType_General, tmp_err );
					}
				
					if( ans->get_boolean() )
					{  // lets run the 'if'-block
						start += 2;
						run_block( tmp_variables, this_private_varlist, ret, ans, start, true );
						if( ret != CodeBlock_SubType_None )
							return;
// 'if' was followed by 'else', lets skip that block
						if( lines.get_size() > start+2 )
							if( lines[start+1]->sub_type == CodeBlock_SubType_Else )
								start = lines[start+2]->pair_with;
					}
					else
					{  // lets skip the 'if'-block
						start = lines[start+1]->pair_with;
// 'if' was followed by 'else', lets run that block
						if( lines.get_size() > start+1 )
						{
							if( lines[start+1]->sub_type == CodeBlock_SubType_Else )
							{
								start += 3; // this skips     }    else    {
								run_block( tmp_variables, this_private_varlist, ret, ans, start, true );
								if( ret != CodeBlock_SubType_None )
									return;
							}
						}
					}
					break;
			
				default:
					break;
				}
			}
			break;

		case CodeBlock_Type_Start:
			start++;
			run_block( tmp_variables, this_private_varlist, ret, ans, start, true );
			if( ret != CodeBlock_SubType_None )
				return;
			break;

		case CodeBlock_Type_Stop:
			if( expect_curley_ending )
				return;
			THROW_ERROR( ErrorType_General, _("Unexpected }") );

		case CodeBlock_Type_Void:
			switch( lines[start]->sub_type )
			{
			case CodeBlock_SubType_Return:
				ans->set_void();
				ret = CodeBlock_SubType_Return;
				return;

			case CodeBlock_SubType_Break:
				ret = CodeBlock_SubType_Break;
				return;

			case CodeBlock_SubType_Continue:
				ret = CodeBlock_SubType_Continue;
				return;

			case CodeBlock_SubType_None:
				break;

			default:
				printf("not yet implemented CodeBlock_Type_Void with sub_type: %ld\n", lines[start]->sub_type);
			}
			break;
		}
	}
}

} // namespace math
