
/*
 *  Math2
 *  Copyright (c) 2003-2007 by Mattias Hultgren <tilda_o_tize@hotmail.com>
 *
 *  See math2.h
 */


#include "math2.h"
#include "math2.intern.h"
#include "keyfile.h"

namespace math
{



enum SymbolTypes { SymbolType_Function, SymbolType_Variable, SymbolType_Other, SymbolType_End };

struct Symbol
{
	utf8_string name;
	SymbolTypes type;
	union{
		struct{
			Variable    *var;
			int32 tmp_index;
		};
		struct{
			BuiltinFunctionPtr func;
			bool need_left, need_right;
			int priority;
		};
	};
	bool hard_linking;
};

void parse_special_cases( Vector<Symbol> &symbols ) throw(error_obj);
void parse_negates( Vector<Symbol> &symbols ) throw(error_obj);
void parse_implicit_multiplication( Vector<Symbol> &symbols ) throw(error_obj);
void parse_absoulute_values( Vector<Symbol> &symbols ) throw(error_obj);

// a 'term' is e.g.   (4+pi)   or   4    or  [[1,2]]  or  sin 4
int get_term_length( const Vector<Symbol> &symbols, int index ) throw(error_obj);


CodeLine::CodeLine()
{
	nroftmpvariables = 0;
}
CodeLine::CodeLine(CodeLine &src) throw(error_obj)
{
	nroftmpvariables = 0;
	*this = src;
}

CodeLine::~CodeLine()
{
	autovarlist.clear();
}


void CodeLine::operator=(CodeLine &src) throw(error_obj)
{
	nodes = src.nodes;
	nroftmpvariables = src.nroftmpvariables;
	autovarlist = src.autovarlist;
	string = src.string;
}

void CodeLine::set_code_line(const char *newstring, bool allow_variable_creation,
                             VariableList *private_varlist) throw(error_obj)
{
	Variable tmpvar;
	Vector<utf8_string> string_symbols;
	Vector<Symbol> symbols;
	Symbol tmp_symbol;
	uint32 idnr;
 
	if(newstring == 0)
		return;

	string = newstring;

	nodes.clear();
	autovarlist.clear();
	nroftmpvariables = 0;

	parse_create_symbol_vector( newstring, &string_symbols );

	if( string_symbols.get_size() == 0 )
		return;
	
	for( int i=0; i<string_symbols.get_size(); i++ )
	{
		tmp_symbol.name = *string_symbols[i];

		if( tmp_symbol.name.get_length() == 1 )
		{
			if( tmp_symbol.name.test_character( 0, "|()[].," ) )
			{
				tmp_symbol.type = SymbolType_Other;
				symbols.append( tmp_symbol );
			}
			else if( tmp_symbol.name.test_character( 0, "&" ) )
			{
				tmp_symbol.name = "and";
				goto CodeLine_symbolizer;
			}
			else
				goto CodeLine_symbolizer;
		}
		else
		{
CodeLine_symbolizer:
			if( tmp_symbol.name.test_character( 0, "0123456789." ) )
			{
				floatx tmp, value;
				Integer int_value = 0;
				bool komma        = false;

				for( int a=0; a<tmp_symbol.name.get_length(); a++ )
				{
					if( tmp_symbol.name.test_character( a, "." ) )
					{
						if( komma )
						{
							THROW_ERROR( ErrorType_General, _("Illegal symbol/character") );
						}
						else
						{
							value = floatx( int_value );
							komma = true;
							tmp = 0.1;
						}
					}
					else
					{
						if( komma )
						{
							value += floatx( tmp_symbol.name.get_digit( a ) ) * tmp;
							tmp /= 10.0;
						}
						else
						{
							int_value *= 10;
							int_value += tmp_symbol.name.get_digit( a );
						}
					}
				}

				if( komma )
					tmpvar.set_complex( Real( value ) );
				else
					tmpvar.set_complex( Real( int_value ) );

				tmp_symbol.type = SymbolType_Variable;
				tmp_symbol.var = autovarlist.create( tmpvar, true, false, false );
				tmp_symbol.hard_linking = true;
				tmp_symbol.tmp_index = -1;

				symbols.append( tmp_symbol );
			}
			else if( tmp_symbol.name.test_character( 0, "\"" ) )
			{
				utf8_string tmp_str;

				tmp_str.assign( tmp_symbol.name, 1, tmp_symbol.name.get_length()-2 );

				tmpvar.set_string( tmp_str );

				tmp_symbol.type = SymbolType_Variable;
				tmp_symbol.var = autovarlist.create( tmpvar, true, false, false );
				tmp_symbol.hard_linking = true;
				tmp_symbol.tmp_index = -1;

				symbols.append( tmp_symbol );
			}
			else
			{
				check_name_against_reserved( tmp_symbol.name );

// check name against private variables
				if( private_varlist )
				{
					try
					{
						tmp_symbol.type = SymbolType_Variable;
						tmp_symbol.tmp_index = int32(private_varlist->get_id(tmp_symbol.name.c_str()));
						private_varlist->get_pointer_rw( uint32(tmp_symbol.tmp_index) );
						tmp_symbol.hard_linking = false;

						symbols.append( tmp_symbol );
						continue;
					}
					catch( ... ) {} // only throwed error is 'Variable not found'
				}

check_for_functions:  // check name against functions
				tmp_symbol.func = function_list.get( tmp_symbol.name, &tmp_symbol.priority, &tmp_symbol.need_left,
				                                     &tmp_symbol.need_right, &tmp_symbol.hard_linking );
				if( tmp_symbol.func != 0 )
				{
					tmp_symbol.type = SymbolType_Function;

					symbols.append( tmp_symbol );
					continue;
				}

// check name against variables
				{
					bool deletable;
					Variable *ptr;

					try{
						idnr = global_varlist.get_id( tmp_symbol.name.c_str() );
						if( (deletable = global_varlist.get_deletable( idnr )) == false )
							ptr = global_varlist.get_pointer_rw( idnr );
					}
					catch(...) // only throwed error is 'Variable not found'
					{
// let's check if an autoloadable file with this name exists
						if( false == tmp_symbol.name.test_character( 0, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_" ) )
							THROW_ERROR( ErrorType_General, _("Illegal symbol/character") );

						if( autoload_file( tmp_symbol.name ) )
							goto check_for_functions;

						if( !allow_variable_creation )
							THROW_ERROR( ErrorType_General, _("Variable/function not found") );

//let's create this Variable then
						tmpvar.set_complex( Complex( Real(0) ) );
						tmpvar.set_name( tmp_symbol.name );
						global_varlist.create( tmpvar, false, true, true );
						tmpvar.set_name( "" );
						deletable = true;
					}

					if( deletable )
						tmp_symbol.hard_linking = false;
					else
					{
						tmp_symbol.hard_linking = true;
						tmp_symbol.var = ptr;
					}
					tmp_symbol.type = SymbolType_Variable;
					tmp_symbol.tmp_index = -1;
					symbols.append( tmp_symbol );
				}
			}
		}
	}

/*	printf( "\nsymbols\n" );
	for( int i=0; i<symbols.get_size(); i++ )
		printf( "  %s\n", symbols[i]->name.c_str() );
	printf( "\n\n" );*/
	parse_absoulute_values( symbols );
/*	printf( "\nsymbols after\n" );
	for( int i=0; i<symbols.get_size(); i++ )
		printf( "  %s\n", symbols[i]->name.c_str() );
	printf( "\n\n" );*/
	parse_special_cases( symbols );
	parse_negates( symbols );
	parse_implicit_multiplication( symbols );

	matrix_bracket_reducing( void_ptr( &symbols ) );
	round_bracket_reducing( void_ptr( &symbols ), 0 );
	create_nodes( void_ptr( &symbols ), 0 );

	parse_pow_mods();
	
	if( nodes.get_size() == 0 )
	{
		if( symbols.get_size() == 1 )
		{
			if( symbols[0]->type == SymbolType_Variable )
			{
				CodeLineNode tmp_node;

				tmp_node.function_name = "set";
				tmp_node.soft_function = false;
				tmp_node.function = builtin_set;
				tmp_node.tmpvariable_index = -1;
				nroftmpvariables = 1;

				tmp_node.left.ptr = 0;
				tmp_node.left.name = "";
				tmp_node.left.soft = false;
				tmp_node.left.tmpvariable_index = 0;

				
				tmp_node.right.name = symbols[0]->name;
				tmp_node.right.tmpvariable_index = -1;
				if( symbols[0]->hard_linking )
				{
					tmp_node.right.ptr = symbols[0]->var;
					tmp_node.right.soft = false;
				}
				else
				{
					tmp_node.right.ptr = 0;
					tmp_node.right.soft = true;
					if( symbols[0]->tmp_index != -1 )
						tmp_node.right.tmpvariable_index = symbols[0]->tmp_index;
				}

				nodes.append( tmp_node ); // creating a tmp copy just to make the variable be copied to answer in calc
			}
		}
	}
}

void CodeLine::parse_pow_mods(void) throw(error_obj)
{
	CodeLineNode *node;
	int i, pow_index=-1;
	
	for( i=0; i<nodes.get_size(); i++ )
	{
		if( nodes[i]->function_name == "^" )
		{
			pow_index = i;
			for( i++; i<nodes.get_size(); i++ )
			{
				if( nodes[i]->function_name == "mod"  &&
				    nodes[i]->left.tmpvariable_index == nodes[pow_index]->tmpvariable_index )
				{
					// this mod is feed from a pow...
			
					// adding a array-call to create the array for powMod
					nodes.insert( *nodes[i], i ); // duplicating the mod...
					node = nodes[i];
					node->soft_function = false;
					node->function = builtin_array_of_two;
					node->function_name = "(array_of_two)";
					node->tmpvariable_index = nodes[pow_index]->tmpvariable_index;
					
					node->left = nodes[pow_index]->right;
					
					//node->right should be left as it is...
					
					
					
					// replacing the mod-call with a powMod call
					node = nodes[i+1];
					node->soft_function = false;
					node->function = builtin_pow_mod;
					node->function_name = "(powMod)";

					node->left = nodes[pow_index]->left;

					node->right.tmpvariable_index = nodes[i]->tmpvariable_index;
					node->right.soft = false;
					node->right.ptr = 0;
					node->right.name = "tmp_var";

					// removing the pow-call
					nodes.remove( pow_index );
					pow_index--; // this makes the for loop recheck this position
					
					break; // jump out of the mod-loop, to search for more pow's
				}
			}
			i = pow_index;
		}
	}
}

void parse_absoulute_values( Vector<Symbol> &symbols ) throw(error_obj)
{
	int i, start, started_round, started_hard;

	while( true )
	{
		for( start=0; ; start++ ) // finds first |
		{
			if( start == symbols.get_size() )
				return;
			if( symbols[start]->type == SymbolType_Other && symbols[start]->name == "|" )
				break;
		}

		while( true )
		{
			for( i=start+1, started_round=0, started_hard=0; ; i++ ) // finds second |
			{
				if( i == symbols.get_size() )
					THROW_ERROR( ErrorType_General, _("Expected closing |") );

				if( started_round != 0 )
				{
					if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "(" )
						started_round++;
					else if( symbols[i]->type == SymbolType_Other && symbols[i]->name == ")" )
						started_round--;
				}
				else if( started_hard != 0 )
				{
					if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "[" )
						started_hard++;
					else if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "]" )
						started_hard--;
				}
				else if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "|" )
					break;
				else if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "(" )
					started_round = 1;
				else if( symbols[i]->type == SymbolType_Other && symbols[i]->name == "[" )
					started_hard = 1;
			}

			if( symbols[i-1]->type != SymbolType_Function  &&
			    !(symbols[i-1]->type == SymbolType_Other && symbols[i-1]->name == "|") )
				break;
			else if( symbols[i-1]->type == SymbolType_Function )
			{
				if( symbols[i-1]->need_right == false )
					break;
			}
			start = i;  // the second | was not a closing, but a starting...
		}

		// replacing the second | with ))
		symbols[i]->name = ")";
		symbols.insert( *symbols[i], i+1 );

		// replacing the first | with (abs(
		symbols[start]->name = "(";
		symbols.insert( *symbols[start], start+1 );
		symbols.insert( *symbols[start], start+1 );

		symbols[start+1]->func = function_list.get( "abs", &symbols[start+1]->priority, &symbols[start+1]->need_left,
				                                   &symbols[start+1]->need_right, &symbols[start+1]->hard_linking );
		symbols[start+1]->type = SymbolType_Function;
		symbols[start+1]->name = "abs";
	}
}

void parse_implicit_multiplication( Vector<Symbol> &symbols ) throw(error_obj)
{
	for( int i=1; i<symbols.get_size(); i++ )
	{
		if( (symbols[i]->type == SymbolType_Other && symbols[i]->name == "(")  ||
		    ((symbols[i]->type == SymbolType_Other && symbols[i]->name == "[")  &&
		     (symbols[i+1]->type == SymbolType_Other && symbols[i+1]->name == "["))  ||
		    symbols[i]->type == SymbolType_Variable  ||
		    (symbols[i]->type == SymbolType_Function  &&  symbols[i]->need_left == false)
		  )
		{
			if( (symbols[i-1]->type == SymbolType_Other && symbols[i-1]->name == ")")  ||
			    (symbols[i-1]->type == SymbolType_Other && symbols[i-1]->name == "]")  ||
			    symbols[i-1]->type == SymbolType_Variable  ||
			    (symbols[i-1]->type == SymbolType_Function  &&  symbols[i-1]->need_right == false)
			  )
			{
				if( symbols[i-1]->type == SymbolType_Other && symbols[i-1]->name == "]" )
				{
					if( i > 2 )
					{
						if( !(symbols[i-2]->type == SymbolType_Other && symbols[i-2]->name == "]") )
							continue;
					}
					else
						continue; // skip this
				}

// inserting a * symbol
				symbols.insert( *symbols[i], i );
				symbols[i]->func = function_list.get( "*", &symbols[i]->priority, &symbols[i]->need_left,
				                                           &symbols[i]->need_right, &symbols[i]->hard_linking );
				symbols[i]->name = "*";
				symbols[i]->type = SymbolType_Function;
			}
		}
	}
}

void parse_negates( Vector<Symbol> &symbols ) throw(error_obj)
{
	for( int i=0; i<symbols.get_size(); i++ ) // search for -(minus) to replace by -(neg)
	{
		if( symbols[i]->type == SymbolType_Function  &&  symbols[i]->name == "-" )
		{
			if( i == 0 )
				goto fix_the_negation;
			if( symbols[i-1]->type == SymbolType_Other &&
			    (symbols[i-1]->name == ","  ||
			     symbols[i-1]->name == "("  ||
			     symbols[i-1]->name == "[" )
			  )
				goto fix_the_negation;
			else
			{
				if( symbols[i-1]->type == SymbolType_Function  &&  symbols[i-1]->need_right )
				{
fix_the_negation:
					int len = get_term_length( symbols, i+1 );

					if( len == 0 )
						THROW_ERROR( ErrorType_General, _("Unexpected end after '-'") );

					symbols[i]->type = SymbolType_Other;
					symbols[i]->name = "(";
					symbols.insert( *symbols[i], i+1 );
					symbols.insert( *symbols[i], i+2 );
					symbols[i+1]->func = function_list.get( "-(neg)", &symbols[i+1]->priority,
					                                                  &symbols[i+1]->need_left,
					                                                  &symbols[i+1]->need_right,
					                                                  &symbols[i+1]->hard_linking );
					symbols[i+1]->type = SymbolType_Function;
					symbols[i+1]->name = "-(neg)";

					symbols.insert( *symbols[i], i+len+3 );
					symbols.insert( *symbols[i], i+len+4 );
					symbols[i+len+3]->name = ")";
					symbols[i+len+4]->name = ")";
				}
			}
		}
	}
}

int get_term_length( const Vector<Symbol> &symbols, int index ) throw(error_obj)
{
	int started_brackets = 1;

	if( symbols[index]->type == SymbolType_Variable )
		return 1;
	else if( symbols[index]->type == SymbolType_Other  &&  symbols[index]->name == "(" )
	{
		for( int i=1; ; i++ )
		{
			if( (index+i) == symbols.get_size() )
			{
				THROW_ERROR( ErrorType_General, _("Expected )") );
			}
			else if( symbols[index]->type == SymbolType_Other  &&  symbols[index+i]->name == "(" )
				started_brackets++;
			else if( symbols[index+i]->type == SymbolType_Other  &&  symbols[index+i]->name == ")" )
			{
				started_brackets--;
				if( started_brackets == 0 )
					return i+1;
			}
		}
	}
	else if( symbols[index]->type == SymbolType_Other  &&  symbols[index]->name == "[" )
	{
		for( int i=1; ; i++ )
		{
			if( (index+i) == symbols.get_size() )
			{
				THROW_ERROR( ErrorType_General, _("Expected ]") );
			}
			else if( symbols[index+i]->type == SymbolType_Other  &&  symbols[index+i]->name == "[" )
				started_brackets++;
			else if( symbols[index+i]->type == SymbolType_Other  &&  symbols[index+i]->name == "]" )
			{
				started_brackets--;
				if( started_brackets == 0 )
					return i+1;
			}
		}
	}
	else if( symbols[index]->type == SymbolType_Function )
	{
		if( symbols[index]->need_right )
			return get_term_length( symbols, index+1 ) +1;
		else
			return 1;
	}

	return 0;
}

void parse_special_cases( Vector<Symbol> &symbols ) throw(error_obj)
{
	for( int i=0; i<(symbols.get_size()-1); i++ )
	{
		if( symbols[i]->type == SymbolType_Variable  &&
		    (symbols[i+1]->type == SymbolType_Other  &&  symbols[i+1]->name == "(") )
		{ // found    ... var(variable_values) ... syntax, let's replace it with a function call
		//  ... ( var.(eval)(variable_values) ) ...
			int end;
			int started = 0;
			for( end=i+2; ; end++ )
			{
				if( end == symbols.get_size() )
					THROW_ERROR( ErrorType_General, _("Expected )") );
				if( symbols[end]->type == SymbolType_Other  &&  symbols[end]->name == "(" )
					started++;
				else if( symbols[end]->type == SymbolType_Other  &&  symbols[end]->name == ")" )
				{
					if( started != 0 )
						started--;
					else
						break;
				}
			}
			end += 3;

			symbols.insert( *symbols[i], i ); // inserting three new symbols
			symbols.insert( *symbols[i], i );
			symbols.insert( *symbols[i], i );

			symbols[i]->type = SymbolType_Other;
			symbols[i]->name = "(";
			symbols[i+2]->type = SymbolType_Other;
			symbols[i+2]->name = ".";
			*symbols[i+4] = *symbols[i];
			symbols[i+3]->type = SymbolType_Function;
			symbols[i+3]->name = "(eval)";
			symbols[i+3]->func = function_list.get( "(eval)", &symbols[i+3]->priority, &symbols[i+3]->need_left,
			                                                  &symbols[i+3]->need_right, &symbols[i+3]->hard_linking );

			symbols.insert( *symbols[i], end );
			symbols[end]->name = ")";
			*symbols[end+1] = *symbols[end];
		}
		else if( symbols[i]->type == SymbolType_Variable  &&
		         (symbols[i+1]->type == SymbolType_Other  &&  symbols[i+1]->name == "[") )
		{ // found    ... var[index_expr] ... syntax
// let's find the end of the [index_expr]
			if( (i+2) >= symbols.get_size() )
				continue;
			if( symbols[i+2]->type == SymbolType_Other  &&  symbols[i+2]->name == "[" )
				continue;

			int end;
			int started = 0;
			for( end=i+2; ; end++ )
			{
				if( end == symbols.get_size() )
					THROW_ERROR( ErrorType_General, _("Expected ]") );
				if( symbols[end]->type == SymbolType_Other  &&  symbols[end]->name == "[" )
					started++;
				else if( symbols[end]->type == SymbolType_Other  &&  symbols[end]->name == "]" )
				{
					if( started != 0 )
						started--;
					else
						break;
				}
			}
			if( (end+1) < symbols.get_size() )
			{
				if( i == 0  &&  symbols[end+1]->type == SymbolType_Function  &&  symbols[end+1]->name == "=" )
				{
// found    var[index_expr] = ... syntax, let's replace with a function call
//  var.[set]((index_expr), ... )
					end += 3;

					symbols.insert( *symbols[i], i );
					symbols.insert( *symbols[i], i );
					symbols.insert( *symbols[i], i );

					symbols[i+1]->type = SymbolType_Other;
					symbols[i+1]->name = ".";
					symbols[i+2]->type = SymbolType_Function;
					symbols[i+2]->name = "[set]";
					symbols[i+2]->func = function_list.get( "[set]", &symbols[i+2]->priority,
					                                                 &symbols[i+2]->need_left,
					                                                 &symbols[i+2]->need_right,
					                                                 &symbols[i+2]->hard_linking );
					symbols[i+3]->type = SymbolType_Other;
					symbols[i+3]->name = "(";
					*symbols[i+4] = *symbols[i+3];

					symbols[end]->name = ")";  // replacing  ]
					symbols[end+1]->name = ",";  // replacing  =
					symbols[end+1]->type = SymbolType_Other;

					symbols.append( *symbols[end] );
				}
				else
					goto the_other_else;
			}
			else
			{
the_other_else:
// found    ... var[index_expr] ... syntax, let's replace with a function call
//  ... ( var.[get](index_expr) ) ...
				end += 3;

				symbols.insert( *symbols[i], i );
				symbols.insert( *symbols[i], i );
				symbols.insert( *symbols[i], i );

				symbols[i]->type = SymbolType_Other;
				symbols[i]->name = "(";
				symbols[i+2]->type = SymbolType_Other;
				symbols[i+2]->name = ".";
				symbols[i+3]->type = SymbolType_Function;
				symbols[i+3]->name = "[get]";
				symbols[i+3]->func = function_list.get( "[get]", &symbols[i+3]->priority, &symbols[i+3]->need_left,
				                                                 &symbols[i+3]->need_right, &symbols[i+3]->hard_linking );
				*symbols[i+4] = *symbols[i];

				symbols.insert( *symbols[i], end );
				symbols[end]->name = ")";
				symbols[end+1]->name = ")";
			}
		}
	}
}

void CodeLine::round_bracket_reducing(void *ptr, int start_at) throw(error_obj)
{
	int i, start;
	bool isArray;
	Array arr;
	Variable tmp_var;
	Vector<Symbol> *symbols = (Vector<Symbol> *)(ptr);
	Symbol end_symbol, tmpvar_symbol;
	CodeLineNode tmp_node;

	end_symbol.type = SymbolType_End;

	tmpvar_symbol.type = SymbolType_Variable;
	tmpvar_symbol.var = 0;
	tmpvar_symbol.hard_linking = true;
	tmpvar_symbol.tmp_index = -1;


	for( i=start_at, start = -1; i<symbols->get_size(); i++ )
	{
rerun_reducer:
		if( (*symbols)[i]->type == SymbolType_End )
			break;
		if( (*symbols)[i]->type == SymbolType_Other )
		{
			if( (*symbols)[i]->name == "(" )
			{
				start = i;
				isArray = false;
			}
			else if( (*symbols)[i]->name == ")" )
			{
				if( start == -1 )
					THROW_ERROR( ErrorType_General, _("Unexpected )") );
				symbols->insert( end_symbol, i );
				if( isArray )
				{
					int32 array_index = nroftmpvariables;

					tmp_node.function_name = "array";
					tmp_node.soft_function = false;
					tmp_node.function = builtin_array;
					tmp_node.tmpvariable_index = nroftmpvariables;
					nroftmpvariables++;

					tmp_node.left.ptr = 0;
					tmp_node.left.name = "";
					tmp_node.left.soft = false;
					tmp_node.left.tmpvariable_index = -1;

					tmp_var.set_complex( Complex( Real(0) ) );
					tmp_node.right.ptr = autovarlist.create( tmp_var, true, false, false );
					tmp_node.right.name = "";
					tmp_node.right.soft = false;
					tmp_node.right.tmpvariable_index = -1;

					nodes.append( tmp_node ); // creating the array

					do
					{
						for( i=start+1; (*symbols)[i]->type != SymbolType_End; i++ )
						{
							if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "," )
								break;
						}
						symbols->insert( end_symbol, i );

						create_nodes( ptr, start+1 );

						tmp_node.function_name = "append";
						tmp_node.soft_function = false;
						tmp_node.function = builtin_append;
						tmp_node.tmpvariable_index = -1;

						tmp_node.left.ptr = 0;
						tmp_node.left.name = "";
						tmp_node.left.soft = false;
						tmp_node.left.tmpvariable_index = array_index;

						if( (*symbols)[start+1]->type != SymbolType_Variable )
							THROW_ERROR( ErrorType_General, _("Syntax error") );
						tmp_node.right.ptr = (*symbols)[start+1]->var;
						tmp_node.right.name = (*symbols)[start+1]->name;
						tmp_node.right.soft = !( (*symbols)[start+1]->hard_linking );
						tmp_node.right.tmpvariable_index = -1;
						tmp_node.right.tmpvariable_index = (*symbols)[start+1]->tmp_index;

						nodes.append( tmp_node ); // appending a variable to the array

						symbols->remove( start+1 ); // removing variable
						symbols->remove( start+1 ); // removing end_symbol
						if( (*symbols)[start+1]->type == SymbolType_Other  &&  (*symbols)[start+1]->name == "," )
							symbols->remove( start+1 ); // removing ,-symbol

					}while( (*symbols)[start+1]->type != SymbolType_End );

					symbols->remove( start ); // removing (-symbol
					symbols->remove( start ); // removing end_symbol

					tmpvar_symbol.tmp_index = array_index;
					*(*symbols)[start] = tmpvar_symbol; // replacing )-symbol with array-tmpvar-symbol
				}
				else
				{
					create_nodes( ptr, start+1 );

					symbols->remove( start );   // removing (-symbol
					symbols->remove( start+1 ); // removing end_symbol
					symbols->remove( start+1 ); // removing )-symbol
				}
				i = start_at;
				start = -1;
				goto rerun_reducer;
			}
			else if( (*symbols)[i]->name == "," )
			{
				if( start == -1 )
					THROW_ERROR( ErrorType_General, _("Separator \',\' found outside brackets.") );

				isArray = true;
			}
		}
	}
	if( start != -1 )
		THROW_ERROR( ErrorType_General, _("Expected )") );
}

int32 CodeLine::parse_matrix_creation( uint32 col, uint32 row ) throw(error_obj)
{
	CodeLineNode tmp_node;
	Variable tmp_var;
	Array tmp_arr;

	tmp_var.set_complex( Complex( Real( int64(row) ) ) );
	tmp_arr.add_variable( &tmp_var );

	tmp_var.set_complex( Complex( Real( int64(col) ) ) );
	tmp_arr.add_variable( &tmp_var );

	tmp_var.set_array( &tmp_arr );


	tmp_node.function_name = "matrix";
	tmp_node.soft_function = false;
	tmp_node.function = builtin_matrix;
	tmp_node.tmpvariable_index = nroftmpvariables;
	nroftmpvariables++;

	tmp_node.left.ptr = 0;
	tmp_node.left.name = "";
	tmp_node.left.soft = false;
	tmp_node.left.tmpvariable_index = -1;

	tmp_node.right.ptr = autovarlist.create( tmp_var, true, false, false );
	tmp_node.right.name = "";
	tmp_node.right.soft = false;
	tmp_node.right.tmpvariable_index = -1;

	nodes.append( tmp_node ); // creating the matrix

	return nroftmpvariables-1;
}


void CodeLine::matrix_bracket_reducing(void *ptr) throw(error_obj)
{
	uint32 col, row, irow, icol;
	int32 i, start, stop, start2, tmp, inside_rounds;
	Vector<Symbol> *symbols = (Vector<Symbol> *)(ptr);
	int32 matrix_tmp_index;
	Symbol end_symbol, tmpvar_symbol;
	CodeLineNode tmp_node;

	end_symbol.type = SymbolType_End;

	tmpvar_symbol.type = SymbolType_Variable;
	tmpvar_symbol.var = 0;
	tmpvar_symbol.hard_linking = true;
	tmpvar_symbol.tmp_index = -1;


	for( i=0; i<symbols->get_size(); i++ )
	{
		if( (*symbols)[i]->type == SymbolType_Other )
		{
			if( (*symbols)[i]->name == "[" )
			{
				start = i;

// lets count the Matrix's rows
				for( row=0, tmp=0, i=start+1;  ; i++ )
				{
					if( i == symbols->get_size() )
					{
						if( tmp == 0 )
						{
							if( row == 0 )
							{
								THROW_ERROR( ErrorType_General, _("Expected [") );
							}
							else
								THROW_ERROR( ErrorType_General, _("Expected ]") );
						}
						else
							THROW_ERROR( ErrorType_General, _("Expected ]") );
					}
					if( tmp == 0 )
					{
						if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "[" )
							tmp = 1;
						else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "]" )
						{
							stop = i;
							if( row == 0 )
								THROW_ERROR( ErrorType_General, _("Expected ]") );
							break;
						}
						else
						{
							if( row == 0 )
							{
								THROW_ERROR( ErrorType_General, _("Expected [") );
							}
							else
								THROW_ERROR( ErrorType_General, _("Expected ]") );
						}
					}
					else
					{
						if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "[" )
						{
							THROW_ERROR( ErrorType_General, _("Unexpected [") );
						}
						else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "]" )
						{
							row++;
							tmp = 0;
						}
					}
				}
// rows counted
// let's count the Matrix's columns
				for( col=1, i=start+1, inside_rounds=0; ; i++ )
				{
					if( i == symbols->get_size() )
					{
						THROW_ERROR( ErrorType_General, _("Expected (") );
					}
					else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "(" )
						inside_rounds++;
					else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == ")" )
					{
						if( !inside_rounds )
							THROW_ERROR( ErrorType_General, _("Unexpected )") );
						inside_rounds--;
					}
					if( !inside_rounds )
					{
						if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "," )
							col++;
						else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "]" )
							break;
					}
				}
// columns counted

				matrix_tmp_index = parse_matrix_creation( col, row );

// calculate the Matrix
				for( irow=1, i=start; irow<=row; irow++ )
				{
					i++;
					if( (*symbols)[i]->type != SymbolType_Other  ||  (*symbols)[i]->name != "[" )
						THROW_ERROR( ErrorType_General, _("Syntax error") );

					for( icol=1; icol<=col; icol++)
					{
						i++;
						start2 = i;
						for( inside_rounds=0; ; i++ )
						{
							if( i == symbols->get_size() )
							{	THROW_ERROR( ErrorType_General, _("Syntax error") ); }
							else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "(" )
								inside_rounds++;
							else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == ")" )
							{
								if( !inside_rounds )
									THROW_ERROR( ErrorType_General, _("Unexpected )") );
								inside_rounds--;
							}
							if( !inside_rounds )
							{

								if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "," )
								{
									if( icol == col )
										THROW_ERROR( ErrorType_General, _("Syntax error") );

									symbols->insert( end_symbol, i );
									break;
								}
								else if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "]" )
								{
									if( icol != col )
										THROW_ERROR( ErrorType_General, _("Syntax error") );

									symbols->insert( end_symbol, i );
									break;
								}
							}
						}

						round_bracket_reducing( ptr, start2 ); // reduce the (if any) round brackets

						create_nodes( ptr, start2 ); // create the work list of this element

						tmp_node.function_name = "[set_next]";
						tmp_node.soft_function = false;
						tmp_node.function = builtin_hook_set_next;
						tmp_node.tmpvariable_index = -1;

						tmp_node.left.ptr = 0;
						tmp_node.left.name = "";
						tmp_node.left.soft = false;
						tmp_node.left.tmpvariable_index = matrix_tmp_index;

						if( (*symbols)[start2]->type != SymbolType_Variable )
							THROW_ERROR( ErrorType_General, _("Syntax error") );
						tmp_node.right.ptr = (*symbols)[start2]->var;
						tmp_node.right.name = (*symbols)[start2]->name;
						tmp_node.right.soft = !( (*symbols)[start2]->hard_linking );
						tmp_node.right.tmpvariable_index = (*symbols)[start2]->tmp_index;

						nodes.append( tmp_node ); // setting next element of the matrix

						symbols->remove( start2 ); // removing variable
						symbols->remove( start2 ); // removing end_symbol
						symbols->remove( start2 ); // removing , or ] symbol
						i = start2-1;
					}
					if( irow == row )
					{
						if( (*symbols)[i+1]->type != SymbolType_Other  ||  (*symbols)[i+1]->name != "]" )
							THROW_ERROR( ErrorType_General, _("Syntax error") );
					}
				}
// Matrix calculated
				while( true )
				{
					if( (*symbols)[start]->type == SymbolType_Other  &&  (*symbols)[start]->name == "]" )
						break;
					if( start >= symbols->get_size() )
						THROW_ERROR( ErrorType_General, _("Syntax error") );
					symbols->remove( start );
				}

				tmpvar_symbol.tmp_index = matrix_tmp_index;
				*(*symbols)[start] = tmpvar_symbol; // replacing ]-symbol with matrix-tmpvar-symbol

				i = -1;
			}
			else if( (*symbols)[i]->name == "]" ) // a Matrix closed that hasn't been opened
				THROW_ERROR( ErrorType_General, _("Unexpected ]") );
		}
	}
}

void CodeLine::create_nodes(void * ptr, int start_at) throw(error_obj)
{
	Vector<Symbol> *symbols = (Vector<Symbol> *)(ptr);
	int32 prio, i;
	CodeLineNode tmp_node;
	Symbol tmpvar_symbol;

	tmpvar_symbol.type = SymbolType_Variable;
	tmpvar_symbol.var = 0;
	tmpvar_symbol.hard_linking = true;
	tmpvar_symbol.tmp_index = -1;

	if( (*symbols)[start_at]->type == SymbolType_End )
		THROW_ERROR( ErrorType_General, _("Empty bracket") );

// Call all object methods
	for( i=start_at; i<symbols->get_size(); i++ )
	{
		if( (*symbols)[i]->type == SymbolType_End )
			break;
		if( (*symbols)[i]->type == SymbolType_Other  &&  (*symbols)[i]->name == "." )
		{
			if( i == start_at )
				THROW_ERROR( ErrorType_General, _("Unexpected . (point)") );
			if( (*symbols)[i-1]->type != SymbolType_Variable )
				THROW_ERROR( ErrorType_General, _("Unexpected . (point)") );
			if( (i+1) == symbols->get_size() )
				THROW_ERROR( ErrorType_General, _("Unexpected end of line") );
			if(  (*symbols)[i+1]->type == SymbolType_End )
				THROW_ERROR( ErrorType_General, _("Unexpected end of line") );

			if( (*symbols)[i+1]->type != SymbolType_Function  ||  (*symbols)[i+1]->priority != 15  ||
			    ((*symbols)[i+1]->need_left) == false )
				THROW_ERROR( ErrorType_General, _("Expected decimals or method name.") );
 
			tmp_node.soft_function = !( (*symbols)[i+1]->hard_linking );
			tmp_node.function = (*symbols)[i+1]->func;
			tmp_node.function_name = (*symbols)[i+1]->name;
			tmp_node.tmpvariable_index = nroftmpvariables;
			nroftmpvariables++;

			if( (*symbols)[i-1]->hard_linking )
			{
				tmp_node.left.tmpvariable_index = (*symbols)[i-1]->tmp_index;
				tmp_node.left.soft = false;
				tmp_node.left.ptr = (*symbols)[i-1]->var;
				tmp_node.left.name = (*symbols)[i-1]->name;
			}
			else
			{
				tmp_node.left.tmpvariable_index = (*symbols)[i-1]->tmp_index;
				tmp_node.left.soft = true;
				tmp_node.left.ptr = 0;
				tmp_node.left.name = (*symbols)[i-1]->name;
			}

			if( (*symbols)[i+1]->need_right )
			{
				if( (i+2) == symbols->get_size() )
					THROW_ERROR( ErrorType_General, _("Syntax error") );
				if(  (*symbols)[i+2]->type == SymbolType_End )
					THROW_ERROR( ErrorType_General, _("Syntax error") );

				if( (*symbols)[i+2]->type != SymbolType_Variable )
					THROW_ERROR( ErrorType_General, _("Syntax error") );

				if( (*symbols)[i+2]->hard_linking )
				{
					tmp_node.right.tmpvariable_index = (*symbols)[i+2]->tmp_index;
					tmp_node.right.soft = false;
					tmp_node.right.ptr = (*symbols)[i+2]->var;
					tmp_node.right.name = (*symbols)[i+2]->name;
				}
				else
				{
					tmp_node.right.tmpvariable_index = (*symbols)[i+2]->tmp_index;
					tmp_node.right.soft = true;
					tmp_node.right.ptr = 0;
					tmp_node.right.name = (*symbols)[i+2]->name;
				}
			}
			else
			{
				tmp_node.right.tmpvariable_index = -1;
				tmp_node.right.soft = false;
				tmp_node.right.ptr = 0;
				tmp_node.right.name = "";
			}

			nodes.append( tmp_node );

			tmpvar_symbol.tmp_index = nroftmpvariables-1;
			*(*symbols)[i-1] = tmpvar_symbol; // replacing left with the result variable

			if( (*symbols)[i+1]->need_right )
				symbols->remove( i+2 ); // right
			symbols->remove( i+1 ); // method
			symbols->remove( i ); // .

			i--; // this setup i for a recheck of the present location in symbols
		}
	}
// All object methods has been added to the node-list


// priority 15 is reserved for object methods
	for( prio=14; prio>=1; prio-- ) // checks the prio 14-1
	{
		for( i=start_at; i<symbols->get_size(); i++ )
		{
			if( (*symbols)[i]->type == SymbolType_End )
				break;

			if( (*symbols)[i]->type == SymbolType_Function  &&  (*symbols)[i]->priority == prio )
			{
				tmp_node.soft_function = !( (*symbols)[i]->hard_linking );
				tmp_node.function = (*symbols)[i]->func;
				tmp_node.function_name = (*symbols)[i]->name;
				tmp_node.tmpvariable_index = nroftmpvariables;
				nroftmpvariables++;

				if( (*symbols)[i]->need_left )
				{
					if( i == start_at )
						THROW_ERROR( ErrorType_General, _("Syntax error") );

					if( (*symbols)[i-1]->type != SymbolType_Variable )
						THROW_ERROR( ErrorType_General, _("Syntax error") );

					if( (*symbols)[i-1]->hard_linking )
					{
						tmp_node.left.tmpvariable_index = (*symbols)[i-1]->tmp_index;
						tmp_node.left.soft = false;
						tmp_node.left.ptr = (*symbols)[i-1]->var;
						tmp_node.left.name = (*symbols)[i-1]->name;
					}
					else
					{
						tmp_node.left.tmpvariable_index = (*symbols)[i-1]->tmp_index;
						tmp_node.left.soft = true;
						tmp_node.left.ptr = 0;
						tmp_node.left.name = (*symbols)[i-1]->name;
					}
				}
				else
				{
					tmp_node.left.tmpvariable_index = -1;
					tmp_node.left.soft = false;
					tmp_node.left.ptr = 0;
					tmp_node.left.name = "";
				}

				if( (*symbols)[i]->need_right )
				{
					if( (i+1) == symbols->get_size() )
						THROW_ERROR( ErrorType_General, _("Syntax error") );
					if(  (*symbols)[i+1]->type == SymbolType_End )
						THROW_ERROR( ErrorType_General, _("Syntax error") );

					if( (*symbols)[i+1]->type != SymbolType_Variable )
						THROW_ERROR( ErrorType_General, _("Syntax error") );

					if( (*symbols)[i+1]->hard_linking )
					{
						tmp_node.right.tmpvariable_index = (*symbols)[i+1]->tmp_index;
						tmp_node.right.soft = false;
						tmp_node.right.ptr = (*symbols)[i+1]->var;
						tmp_node.right.name = (*symbols)[i+1]->name;
					}
					else
					{
						tmp_node.right.tmpvariable_index = (*symbols)[i+1]->tmp_index;
						tmp_node.right.soft = true;
						tmp_node.right.ptr = 0;
						tmp_node.right.name = (*symbols)[i+1]->name;
					}
				}
				else
				{
					tmp_node.right.tmpvariable_index = -1;
					tmp_node.right.soft = false;
					tmp_node.right.ptr = 0;
					tmp_node.right.name = "";
				}

				nodes.append( tmp_node );

				tmpvar_symbol.tmp_index = nroftmpvariables-1;

				if( (*symbols)[i]->need_right )
					symbols->remove( i+1 ); // right

				if( (*symbols)[i]->need_left )
				{
					*(*symbols)[i-1] = tmpvar_symbol; // replacing left with the result variable
					symbols->remove( i ); // function
				}
				else
					*(*symbols)[i] = tmpvar_symbol; // replacing function with the result variable

				i--; // this setup i for a recheck of the present location in symbols
			}
		}
	}

	if( (start_at+1) != symbols->get_size() )
	{
		if( (*symbols)[start_at+1]->type != SymbolType_End )
			THROW_ERROR( ErrorType_General, _("Syntax error") );
	}
}

void CodeLine::calc( Variable *answer, VariableList *private_varlist, Vector<Variable> *tmp_variables ) throw(error_obj)
{
	Vector<Variable> tmpvarlist;
	Variable *left, *right, *res;
	CodeLineNode *node;
	int i;

	if( nodes.get_size() == 0 )
		THROW_ERROR( ErrorType_General, _("Nothing to calculate") );

	if( tmp_variables == 0 )
		tmp_variables = &tmpvarlist;

	if( tmp_variables->get_size() < nroftmpvariables )
		tmp_variables->set_size_filled( nroftmpvariables );

	try
	{
		// calculate everything here
		for( i=0; i<nodes.get_size(); i++ )
		{
			node = nodes[i];
			if( node->left.soft )
			{
				if( node->left.tmpvariable_index != -1 )
					left = private_varlist->get_pointer_rw( uint32(node->left.tmpvariable_index) );
				else
					left = global_varlist.get_pointer_rw( global_varlist.get_id( node->left.name.c_str() ) );
			}
			else if( node->left.tmpvariable_index != -1 )
				left = (*tmp_variables)[ node->left.tmpvariable_index ];
			else
				left = node->left.ptr;

			if( node->right.soft )
			{
				if( node->right.tmpvariable_index != -1 )
					right = private_varlist->get_pointer_rw( uint32(node->right.tmpvariable_index) );
				else
					right = global_varlist.get_pointer_rw( global_varlist.get_id( node->right.name.c_str() ) );
			}
			else if( node->right.tmpvariable_index != -1 )
				right = (*tmp_variables)[ node->right.tmpvariable_index ];
			else
				right = node->right.ptr;

			if( node->tmpvariable_index != -1 )
				res = (*tmp_variables)[node->tmpvariable_index];
			else
				res = 0;
			if( node->soft_function )
				CodeBlockFunctionPtr(node->function)( res, left, right, node->function_name );
			else
				node->function( res, left, right, private_varlist );
		}
		if( answer != 0 )
			*answer = *(*tmp_variables)[nroftmpvariables-1];
	}
	catch( ... )
	{
		global_varlist.set_all_to_unused();
		throw;
	}
}

} // namespace math
