
/*
 *  Math2
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See math2.h
 */


#include "math2.h"
#include "keyfile.h"
#include <string.h>
#include <stdio.h>
#include <new>

namespace math
{

void Variable::append_to_string( utf8_string &str, const Format &fmt, bool show_name ) const
       throw(error_obj)
{
	utf8_string tmp_str;

	try
	{
		if( has_name()  &&  show_name )
		{
			str.append( *get_name() );
			tmp_str = " = ";
			str.append( tmp_str );
		}

		switch( get_type() )
		{
		case VariableType_Matrix:
			get_matrix()->append_to_string( str, fmt, (has_name()  &&  show_name) );
			break;

		case VariableType_String:
			tmp_str = "\"";
			str.append( tmp_str );
			str.append( *get_string() );
			str.append( tmp_str );
			break;

		case VariableType_Complex:
			get_complex()->append_to_string( str, fmt );
			break;

		case VariableType_Picture:
			{
				const Picture *pic = get_picture();
				char tmp[100];
				snprintf( tmp, 100, "%s %ldx%ld", _("Picture"), pic->pic.get_width(),
				          pic->pic.get_height() );
				tmp_str = tmp;
				str.append( tmp_str );
			}
			break;

		case VariableType_Array:
			{
				const Array *array = get_array();

				if( array->get_size() > 0 )
				{
					tmp_str = "( ";
					str.append( tmp_str );
					if( array->get_variable(0)->get_type() == VariableType_Complex )
						array->get_variable(0)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(0)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(0)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(0)->get_type()];
						str.append( tmp_str );
					}
				}
				else
				{
					tmp_str = "( ";
					str.append( tmp_str );
					tmp_str = _("Empty array");
					str.append( tmp_str );
				}

				for( uint32 i=1; i<array->get_size(); i++ )
				{
					tmp_str = ", ";
					str.append( tmp_str );
					if( array->get_variable(i)->get_type() == VariableType_Complex )
						array->get_variable(i)->get_complex()->append_to_string( str, fmt );
					else if( array->get_variable(i)->get_type() == VariableType_Boolean )
					{
						tmp_str = (array->get_variable(i)->get_boolean() ? "true" : "false");
						str.append( tmp_str );
					}
					else
					{
						tmp_str = VARIABLE_TYPE_NAMES[array->get_variable(i)->get_type()];
						str.append( tmp_str );
					}
				}

				tmp_str = " )";
				str.append( tmp_str );
			}
			break;

		case VariableType_Boolean:
			tmp_str = (get_boolean() ? "true" : "false");
			str.append( tmp_str );
			break;

		case VariableType_Void:
			str.append( "void" );
			break;
		}
	}
	catch(error_obj error) {  throw error;                 }
	catch(...)             {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}


Variable::Variable()
{
	type = VariableType_Boolean;

	isconstant = false;
	can_change_type = true;

	inuse = false;

	boolean = true;

	change_func = 0;
	delete_func = 0;
}
Variable::Variable(const Variable &src) throw(error_obj)
{
	type = VariableType_Boolean;

	isconstant = false;
	can_change_type = true;

	inuse = false;

	boolean = true;

	change_func = 0;
	delete_func = 0;

	*this = src;
}

Variable::~Variable()
{
	if( delete_func != 0 )
		(*delete_func)(this);
	change_func = 0;
	isconstant = false;
	clear();
}

// this function only copies the value/Matrix/function/list/string of the Variable
void Variable::operator=(const Variable &src) throw(error_obj)
{
	if(&src == this)
		return;

	switch(src.type)
	{
	case VariableType_Boolean:
		set_boolean(src.boolean);
		break;

	case VariableType_Complex:
		set_complex( *Complex_ptr(src.complex_buf) );
		break;

	case VariableType_Matrix:
		set_matrix(src.matr);
		break;

	case VariableType_String:
		if(src.str == 0)
			set_string("");
		else
		{
			if(src.code_line != 0)
				set_code_line( src.str->c_str() );
			else
				set_string( *src.str );
		}
		break;

	case VariableType_Array:
		set_array(src.array);
		break;

	case VariableType_Picture:
		set_picture(src.pic);
		break;

	case VariableType_Void:
		set_void();
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		{
			char tmp_err[ERROR_OBJ_MSG_LEN];
			snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Data type is unsupported by '%s'"), "Variable::operator=" );
			THROW_ERROR( ErrorType_General, tmp_err );
		}
	}
}

bool Variable::operator==( const Variable &var ) const
{
	if( type != var.type )
		return false;

	else if( type == VariableType_Complex )
		return (*get_complex() == *var.get_complex());

	else if( type == VariableType_Boolean )
		return (get_boolean() == var.get_boolean());

	else if( type == VariableType_Matrix )
		return (*get_matrix() == *var.get_matrix());

	else if( type == VariableType_String )
		return (*get_string() == *var.get_string());
	
	else if( type == VariableType_Picture )
		return (*get_picture() == *var.get_picture());
	
	else  if( type == VariableType_Array )
		return (*get_array() == *var.get_array());

	else // if( type == VariableType_Void )
		return var.is_void();
}

void Variable::set_to_constant(void)
{
	isconstant = true;
}

void Variable::set_can_change_type(bool boolean) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	can_change_type = boolean;
}

void Variable::set_name(const utf8_string &newname) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	try{  name = newname;  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}

const Matrix* Variable::get_matrix(void) const throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Matrix") );

	if(matr == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return matr;
}

Matrix* Variable::get_matrix_rw(void) throw(error_obj)
{
	if(type != VariableType_Matrix)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Matrix") );

	if(matr == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	return matr;
}

const Array* Variable::get_array(void) const throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Array") );

	if(array == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return array;
}

Array* Variable::get_array_rw(void) throw(error_obj)
{
	if(type != VariableType_Array)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Array") );

	if(array == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	return array;
}

const Picture* Variable::get_picture(void) const throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Picture") );

	if(pic == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return pic;
}

Picture* Variable::get_picture_rw(void) throw(error_obj)
{
	if(type != VariableType_Picture)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Picture") );

	if(pic == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	return pic;
}

void Variable::clear(void) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	switch(type)
	{
	case VariableType_Boolean:
		boolean = true;
		break;

	case VariableType_Complex:
		if( boolean )
			Complex_ptr( complex_buf )->~Complex();
		boolean = false;
		break;

	case VariableType_String:
		delete code_line;
		delete str;
		code_line = 0;
		str = 0;
		break;

	case VariableType_Matrix:
		delete matr;
		matr = 0;
		break;

	case VariableType_Array:
		delete array;
		array = 0;
		break;

	case VariableType_Picture:
		delete pic;
		pic = 0;
		break;

	case VariableType_Void:
		break;

	default: // this should only occur when I've forgotten to add a curtain Variable-type to this function
		{
			char tmp_err[ERROR_OBJ_MSG_LEN];
			snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Data type is unsupported by '%s'"), "Variable::clear" );
			THROW_ERROR( ErrorType_General, tmp_err );
		}
	}
	if(change_func != 0)
		(*change_func)(this);
}

void Variable::signal_changed(void)
{
	if( change_func != 0 )
		(*change_func)( this );
}

void Variable::set_complex(const Complex &newvalue) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	if(type != VariableType_Complex)
	{
		if(can_change_type == false)
			THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

		void (*tmp)(Variable *ptr);
		tmp = change_func;
		change_func = 0;

		clear();

		type = VariableType_Complex;
		boolean = false;

		change_func = tmp;
	}

	if( !boolean )
	{
		try{  new (complex_buf) Complex;  boolean = true;  }
		catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
	}

	*Complex_ptr(complex_buf) = newvalue;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_boolean(bool newbool) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	if( can_change_type == false  &&  type != VariableType_Boolean )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	clear();

	type  = VariableType_Boolean;
	boolean = newbool;

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_array(const Array *newarray) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	if( can_change_type == false  &&  type != VariableType_Array )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	if(type == VariableType_Array)
	{
		if(array == newarray) // the array's are the same
			return;
	}
	clear();

	type = VariableType_Array;
	array = 0;

	if(newarray == 0)
		return;

	try{  array = new Array;  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	try{  *array = *newarray;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_matrix(const Matrix *newmatrix) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	if( can_change_type == false  &&  type != VariableType_Matrix )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	if(type == VariableType_Matrix)
	{
		if(matr == newmatrix) // the Matrix' are the same
			return;
	}
	else
	{
		clear();

		type  = VariableType_Matrix;
		matr = 0;

		if(newmatrix == 0)
			return;

		try{  matr = new Matrix;  }
		catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
	}

	try{  *matr = *newmatrix;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_picture(const Picture *newpic) throw(error_obj)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	if( can_change_type == false  &&  type != VariableType_Picture )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	if(type == VariableType_Picture)
	{
		if(pic == newpic) // the Picture's are the same
			return;
	}
	else
	{
		clear();

		type = VariableType_Picture;
		pic = 0;

		if(newpic == 0)
			return;

		try{  pic = new Picture;  }
		catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
	}

	try{  *pic = *newpic;  }
	catch(...)
	{
		clear();
		throw;
	}

	if(change_func != 0)
		(*change_func)(this);
}

const utf8_string* Variable::get_string(void) const throw(error_obj)
{
	if(type != VariableType_String)
		THROW_ERROR( ErrorType_General, _("Variable type isn't String") );

	if(str == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return str;
}

utf8_string* Variable::get_string_rw(void) throw(error_obj)
{
	if(type != VariableType_String)
		THROW_ERROR( ErrorType_General, _("Variable type isn't String") );

	if(str == 0)
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	return str;
}

const utf8_string* Variable::get_name(void) const throw(error_obj)
{
	if( name.get_length() == 0 )
		THROW_ERROR( ErrorType_General, _("Variable has no name.") );

	return &name;
}

void Variable::calc_code_line( Variable *res ) throw(error_obj)
{
	bool tmp_isconstant;

	if(type != VariableType_String)
		THROW_ERROR( ErrorType_General, _("Variable isn't String") );

	if(code_line == 0)
	{
		if(str == 0)
			THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

		try{  code_line = new CodeLine;  }
		catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

		code_line->set_code_line( str->c_str() );
	}

	if(inuse)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Circle reference of variable %s"), name.c_str() );
		THROW_ERROR( ErrorType_General, tmp_err );
	}

	if( has_name() ) // a nameless Variable can't have a circle reference
		inuse = true;
	tmp_isconstant = isconstant;
	isconstant = true;

	try{  code_line->calc( res );  }
	catch(...)
	{
		inuse = false;
		isconstant = tmp_isconstant;
		throw;
	}
	inuse = false;
	isconstant = tmp_isconstant;
}

void Variable::set_string(const utf8_string &newstring) throw(error_obj)
{
	if( isconstant == true )
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );
	if( can_change_type == false  &&  type != VariableType_String )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	clear();

	type = VariableType_String;
	str = 0;
	code_line = 0;

	try{  str = new utf8_string(newstring);  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	if(change_func != 0)
		(*change_func)(this);
}

void Variable::set_code_line(const char *new_code_line) throw(error_obj)
{
	void (*tmp_change_func)(Variable *ptr);

	tmp_change_func = change_func;
	change_func = 0;

	set_string( new_code_line );

	change_func = tmp_change_func;

	code_line = 0;

	if(new_code_line == 0)
		return;
	if(new_code_line[0] == 0)
		return;

	try{  code_line = new CodeLine;  }
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }


	code_line->set_code_line(new_code_line);


	if(change_func != 0)
		(*change_func)(this);
}

const Complex* Variable::get_complex(void) const throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Complex") );

	if( !boolean )
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return Complex_ptr(complex_buf);
}

Complex* Variable::get_complex_rw(void) throw(error_obj)
{
	if(type != VariableType_Complex)
		THROW_ERROR( ErrorType_General, _("Variable type isn't Complex") );

	if( !boolean )
		THROW_ERROR( ErrorType_Internal, "Variable has a NULL-object" );

	return Complex_ptr(complex_buf);
}

bool Variable::get_boolean(void) const throw(error_obj)
{
	if(type == VariableType_Boolean)
		return boolean;

	THROW_ERROR( ErrorType_General, _("Variable type isn't Boolean") );
}

void Variable::set_void(void)
{
	if(isconstant == true)
		THROW_ERROR( ErrorType_General, _("Tried to change a constant.") );

	if( can_change_type == false  &&  type != VariableType_Void )
		THROW_ERROR( ErrorType_General, _("Tried to change type of a type locked variable.") );

	void (*tmp)(Variable *ptr);
	tmp = change_func;
	change_func = 0;

	clear();

	type = VariableType_Void;

	change_func = tmp;

	if(change_func != 0)
		(*change_func)(this);
}

} // namespace math
