
/*
 *  Math2
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See math2.h
 */


#include "math2.h"
#include "math2.intern.h"
#include "keyfile.h"

namespace math
{

VariableList::intern_variablelist::intern_variablelist()
{
	var = 0;
}
VariableList::intern_variablelist::~intern_variablelist()
{
	delete var;
	delete next;
}

VariableList::VariableList()
{
	varlist = 0;
}
VariableList::VariableList(const VariableList &src) throw(error_obj)
{
	varlist = 0;
	*this = src;
}

VariableList::~VariableList()
{
	clear();
}

void VariableList::operator=(const VariableList &src) throw(error_obj)
{
	intern_variablelist *src_first = src.varlist, *first;

	clear();

	while( src_first != 0 )
	{
		create( *(src_first->var), src_first->var->get_isconstant(),
		        src_first->isdeletable, src_first->var->get_can_change_type() );

		src_first = src_first->next;
	}

	src_first = src.varlist;
	first = varlist;
	while( src_first != 0 )
	{
		first->idnr = src_first->idnr;

		src_first = src_first->next;
		first = first->next;
	}
}

void VariableList::clear(void)
{
	delete varlist;
	varlist = 0;
}

uint32 VariableList::get_id(const char *name) const throw(error_obj)
{
	intern_variablelist *first = varlist;

	for(;;)
	{
		if(first == 0)
			THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );

		if(first->var->has_name())
		{
			if(*(first->var->get_name()) == name)
				return first->idnr;
		}

		first = first->next; // setup for next Variablename check
	}
}

bool VariableList::check_for( const utf8_string &name ) const
{
	intern_variablelist *first = varlist;

	for(;;)
	{
		if(first == 0)
			return false;

		if(first->var->has_name())
		{
			if(*(first->var->get_name()) == name)
				return true;
		}

		first = first->next; // setup for next Variablename check
	}
}

Variable * VariableList::get_pointer_rw(uint32 idnr) throw(error_obj)
{
	intern_variablelist *first = varlist;

	for(; first!=0 ; first=first->next )
	{
		if( first->idnr == idnr )
			return first->var;
	}
	THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );
}

const Variable * VariableList::get_pointer(uint32 idnr) const throw(error_obj)
{
	intern_variablelist *first = varlist;

	for(; first!=0 ; first=first->next )
	{
		if( first->idnr == idnr )
			return first->var;
	}
	THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );
}

uint32 VariableList::get_highest_idnr(void) const
{
	intern_variablelist *first = varlist;
	uint32 highest = 0;

	for(; first!=0 ; first=first->next )
	{
		if(first->idnr > highest)
			highest = first->idnr;
	}
	return highest;
}

void VariableList::make_undeletable(uint32 idnr) throw(error_obj)
{
	intern_variablelist *first = varlist;

	for(; first!=0 ; first=first->next )
	{
		if(first->idnr == idnr)
		{
			first->isdeletable = false;
			return;
		}
	}
	THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );
}

bool VariableList::get_deletable(uint32 idnr) const throw(error_obj)
{
	intern_variablelist *first = varlist;

	for(; first!=0 ; first=first->next )
	{
		if(first->idnr == idnr)
			return first->isdeletable;
	}
	THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );
}

void VariableList::remove(uint32 idnr) throw(error_obj)
{
	intern_variablelist *first = varlist;
	intern_variablelist *prev = 0;
	intern_variablelist *tmp;

	tmp = first;
	for(; first!=0 ; first=first->next )
	{
		if(first->idnr == idnr)
		{
			if( first->isdeletable )
			{
				if(first->var != 0)
				{
					if(first->var->get_inuse())
						THROW_ERROR( ErrorType_General, _("Can't delete variable, because it's in use.") );
				}
				if(prev != 0)
					prev->next = first->next; // makes the linked list exclude this Variable
				else
					tmp = first->next;

				first->next = 0;
				delete first;

				varlist = tmp;
				return;
			}
			else
				THROW_ERROR( ErrorType_General, _("Variable is undeletable") );
		}

		prev = first;
	}
	THROW_ERROR( ErrorType_Variable_not_found, _("Variable not found") );
}

void VariableList::set_all_to_unused(void)
{
	intern_variablelist *first = varlist;

	for(; first!=0 ; first=first->next )
		first->var->set_to_unused();
}

void VariableList::add_this(Variable *newvar, bool isconstant, bool isdeletable,
                            bool can_change_type)  throw(error_obj)
{
	intern_variablelist *first = varlist;
	intern_variablelist *tmp;

	if( newvar == 0 ) // no Variable to add
		return;

	if( newvar->has_name() )
	{
		check_name_against_reserved( *(newvar->get_name()) );

		try{
			get_id( newvar->get_name()->c_str() );
			{
				error_obj error;
				error.type = ErrorType_Name_isnt_unique;
				snprintf( error.msg, ERROR_OBJ_MSG_LEN, _("Variable name (%s) is not unique"), newvar->get_name()->c_str() );
				throw error;
			}
		}
		catch(error_obj error)
		{
			if( error.type == ErrorType_Name_isnt_unique )
				throw;
		}

		if( function_list.check_for( *(newvar->get_name()) ) )
		{
			error_obj error;
			error.type = ErrorType_Name_isnt_unique;
			snprintf( error.msg, ERROR_OBJ_MSG_LEN, _("Variable name (%s) is not unique"), newvar->get_name()->c_str() );
			throw error;
		}
	}

	try
	{
		if( first == 0 )
		{
			tmp = new intern_variablelist;

			tmp->next = 0;
			first = tmp;

			tmp->idnr = 1;
		}
		else
		{
			uint32 highestidnr;

			tmp = first;

			for(;tmp->next != 0;) // finding the last element
				tmp = tmp->next;

			highestidnr = tmp->idnr;

			tmp->next = new intern_variablelist; // creating new element after the last one

			tmp = tmp->next;

			tmp->next = 0;

			tmp->idnr = highestidnr + 1;
		}
	}
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	tmp->var = newvar;

	tmp->isdeletable = isdeletable;
	tmp->var->set_can_change_type( can_change_type );

	if( isconstant )
		tmp->var->set_to_constant();

	varlist = first;
}

Variable * VariableList::create(const Variable &newvar, bool isconstant, bool isdeletable,
                                bool can_change_type) throw(error_obj)
{
	Variable *var;
	try{
		var = new Variable;
	}
	catch(...) {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }

	*var = newvar;

	if(newvar.has_name())
	{
		try{  var->set_name( *(newvar.get_name()) );  }
		catch(error_obj error)
		{
			delete var;
			throw error;
		}
	}

	add_this( var, isconstant, isdeletable, can_change_type );

	return var;
}


} // namespace math
