
/*
 *  Matrix - Implements a matrix
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See matrix.h
 */

/*
News
----

v4  2006-07-30  &  2006-08-10
--

	Renamed Matrix::adjunt  to  Matrix::adjoint
	Replaced all throw_... calls with THROW_ERROR
	Replaced all std::string usage with utf8_string

v3  2006-05-21
--

	Small clean up in Matrix::append_to_string

v2  2006-01-08  &  2006-04-01
--

	Fixed a bug in Matrix::identity
	Added function: set_next

v1  2005-10-01 - 2005-11-06
--

	Extracted class Matrix from the math2.* system
	Small fix in append_to_string

*/

#include "matrix.h"
#include "vector.h"
#include "keyfile.h"

void Matrix::append_to_string( utf8_string &str, const Format &fmt, bool row_break ) const
               throw(error_obj)
{
	Vector<utf8_string> vect;
	Vector<uint32> vect_width;
	uint32 x,y,len,tmp;
	utf8_string tmp_str;

	try
	{
		if( width | height ) // if( width != 0  ||  height != 0 )
		{
			for( x=1; x<=width; x++ )
			{
				len = 0;
				for( y=1; y<=height; y++ )
				{
					tmp_str = "";
					get_complex(x,y).append_to_string( tmp_str, fmt );

					vect.append( tmp_str );
					tmp = tmp_str.get_length();
					len = (len < tmp) ? tmp : len;
				}
				vect_width.append( len );
			}

			if( height > 1  &&  row_break )
				str.append( "\n" );

			for( y=1; y<=height; y++ )
			{
				if( y == 1 )
					str.append( "[[  " );
				else
					str.append( " [  " );

				for( x=1; x<=width; x++ )
				{
					tmp_str = " ";
					for( uint32 i=vect[(x-1) * height + y-1]->get_length(); i<*vect_width[x-1] ; i++ )
						str.append( tmp_str );
					str.append( *vect[(x-1) * height + y-1] );
					if( x != width )
						str.append( ", " );
					else
						str.append( "  " );
				}

				if( y != height )
					str.append( "]\n" );
				else
					str.append( "]]" );
			}
		}
		else
			str.append( _("Non-existing matrix") );
	}
	catch(error_obj error) {  throw;  }
	catch(...)             {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}

Matrix::Matrix()
{
	width   = 0;
	height  = 0;
	data    = 0;
	next = 0;
}

Matrix::Matrix(const Matrix &src) throw(error_obj)
{
	width   = 0;
	height  = 0;
	data    = 0;
	next = 0;
	*this = src;
}

Matrix::~Matrix()
{
	delete [] data;
}

void Matrix::set_size(uint32 newwidth, uint32 newheight) throw(error_obj)
{
	if(  newwidth == 0  ||  newheight == 0  )
	{
		width  = 0;
		height = 0;

		delete [] data;
		data = 0;
		return;
	}

	if( (width * height) == (newwidth * newheight) )
	{
		width = newwidth;
		height = newheight;
		return;
	}

	width = newwidth;
	height = newheight;

	delete [] data;

	try{  data = new Complex [width * height];  }
	catch(...)
	{
		width = 0;
		height = 0;
		data = 0;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
}

void Matrix::set_complex(uint32 col,uint32 row,const Complex &value) throw(error_obj)
{
	if( col > 0  &&  row > 0  &&  col <= width  &&  row <= height )
		data[ (col-1) + (row-1) * width ] = value;
}

Complex Matrix::get_complex(uint32 col,uint32 row) const throw(error_obj)
{
	if( col > 0  &&  row > 0  &&  col <= width  &&  row <= height )
		return data[ (col-1) + (row-1) * width ];
	else
		return Complex( Real(0) );
}

void Matrix::trans(void) throw(error_obj)
{
	Matrix tmp;
	uint32 row,col;

	tmp = *this;

	set_size(height,width);

	for(row=1;row<=tmp.height;row++)
	{
		for(col=1;col<=tmp.width;col++)
			set_complex(row,col,tmp.get_complex(col,row));
	}
}

void Matrix::adjoint(void) throw(error_obj)
{
	Matrix tmp,res;
	uint32 row,col,r,c,tr,tc;

	if( width != height  ||  width <= 1 )
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	tmp.set_size( width-1, width-1 );
	res.set_size( width, width );

	for( row=1; row<=width; row++ )
	{
		for( col=1; col<=width; col++ )
		{

// extracting the 'element det'
			for( tr=0,r=1; r<=width; r++ )
			{
				if( r != row )
				{
					if( r > row )
						tr = 1;
					for( tc=0,c=1; c<=width; c++ )
					{
						if( c > col )
							tc = 1;

						if( c != col )
							tmp.set_complex( c-tc, r-tr, get_complex(c,r) );
					}
				}
			}
// extracting complete

			if( (row+col) & 1 )
				res.set_complex( row, col, -tmp.det() );
			else
				res.set_complex( row, col, tmp.det() );
		}
	}
	*this = res;
}

void matrix_add(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj)
{
	uint32 i,i2;

	if( src1->width != src2->width  ||  src1->height != src2->height  ||  src1->width == 0  ||
	    src1->height == 0 )
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	dest->set_size(src1->width,src1->height);

	for(i=0;i<src1->width;i++)
	{
		for(i2=0;i2<src1->height;i2++)
			dest->data[i+i2*(src1->width)] = src1->data[i+i2*(src1->width)] + src2->data[i+i2*(src1->width)];
	}
}

void matrix_sub(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj)
{
	uint32 i,i2;

	if((src1->width != src2->width) | (src1->height != src2->height) | (src1->width == 0) | (src1->height == 0))
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	dest->set_size(src1->width,src1->height);

	for(i=0;i<src1->width;i++)
	{
		for(i2=0;i2<src1->height;i2++)
			dest->data[i+i2*(src1->width)] = src1->data[i+i2*(src1->width)] - src2->data[i+i2*(src1->width)];
	}
}

void Matrix::mul( const Complex &val ) throw(error_obj)
{
	uint32 i,i2;
	Complex *ptr;

	if( width == 0  ||  height == 0 )
		return;

	for( i=0; i<width; i++ )
	{
		for( i2=0; i2<height; i2++ )
		{
			ptr = &data[ i + i2*(width) ];
			*ptr = *ptr * val;
		}
	}
}

void matrix_mul(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj)
{
	uint32 x,y,z;
	Complex tmp;

	if((src1->width != src2->height) | (src1->width == 0) | (src1->height == 0) | (src2->width == 0))
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	dest->set_size(src2->width,src1->height);

	for(y=0;y<dest->height;y++)
	{
		for(x=0;x<dest->width;x++)
		{
			tmp = Complex( Real(0) );
			for(z=0;z<src1->width;z++)
				tmp = tmp + src1->data[z + y * src1->width] * src2->data[x + z * src2->width];

			dest->data[x + y * dest->width] = tmp;
		}
	}
}

void Matrix::fill(const Complex &val) throw(error_obj)
{
	uint32 i;

	for(i=0;i<(width*height);i++)
		data[i] = val;
}

void Matrix::identity(void) throw(error_obj)
{
	Complex val( Real(0) );

	fill(val);

	val.real = Real(1);

	uint32 max = (width < height) ? width : height;
	for( uint32 i=0; i<max; i++ )
		data[ i + i*width ] = val;
}

void Matrix::replace_row(uint32 dest_row, const Matrix *src, uint32 src_row) throw(error_obj)
{
	uint32 i;

	if( width != src->width )
		THROW_ERROR( ErrorType_General, _("Invalid matrix width") );
	if( dest_row > height  ||  dest_row == 0  ||  src_row > src->height  ||  src_row == 0 )
		return;

	for(i=1;i<=width;i++)
		set_complex( i, dest_row, src->get_complex( i, src_row ) );
}

void Matrix::row_swap(uint32 row1,uint32 row2) throw(error_obj)
{
	uint32 i;
	Complex tmp;

	if((row1 > height) | (row1 == 0) | (row2 > height) | (row2 == 0) | (row1 == row2))
		return;

	for(i=0;i<width;i++)
	{
		tmp = data[(row2-1)*width + i];
		data[(row2-1)*width + i] = data[(row1-1)*width + i];
		data[(row1-1)*width + i] = tmp;
	}
}

void Matrix::row_mul(uint32 row,const Complex &val) throw(error_obj)
{
	uint32 i;

	if( row > height  ||  row == 0 )
		return;

	for(i=0;i<width;i++)
		data[(row-1)*width + i] = data[(row-1)*width + i] * val;
}

void Matrix::row_mul_add(uint32 src_row,const Complex &val,uint32 dest_row) throw(error_obj)
{
	uint32 i;

	if((src_row > height) | (src_row == 0) | (dest_row > height) | (dest_row == 0) | (src_row == dest_row))
		return;

	for(i=0;i<width;i++)
		data[(dest_row-1)*width + i] = data[(dest_row-1)*width + i] + data[(src_row-1)*width + i] * val;
}

void Matrix::inverse(void) throw(error_obj)
{
	Complex tmp;

	tmp = det();
	if( tmp == Complex( Real(0) ) )
		THROW_ERROR( ErrorType_General, _("Can't calculate inverse-matrix (determinant equals to zero)") );

	adjoint();

	tmp = Complex( Real(1) ) / tmp;

	mul( tmp );
}

void matrix_pow( Matrix *dest, const Matrix *src, Integer power ) throw(error_obj)
{
	if( dest == 0  ||  src == 0 )
		return;

	if( src->width != src->height  || src->width == 0 )
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	Matrix tmp, tmp2;
	bool invers;

	tmp = *src;

	if( power.isNegative() )
	{
		invers = true;
		power.negate();
	}
	else
		invers = false;

	if( power.check_bit( 0 ) )
		*dest = tmp;
	else
	{
		dest->set_size( tmp.width, tmp.height );
		dest->identity();
	}

	power.div_by_2();
	while( power != 0 )
	{
		matrix_mul( &tmp2, &tmp, &tmp );
		tmp = tmp2;

		if( power.check_bit( 0 ) )
		{
			matrix_mul( &tmp2, &tmp, dest ); 
			*dest = tmp2;
		}
		power.div_by_2();
	}

	if( invers )
		dest->inverse();
}

void Matrix::operator=(const Matrix &src) throw(error_obj)
{
	next = src.next;
	set_size(src.width,src.height);
	for(uint32 i=0;i<(width*height);i++)
		data[i] = src.data[i];
}

Complex Matrix::det(void) const throw(error_obj)
{
	if( width != height  ||  width == 0  ||  height == 0 )
		THROW_ERROR( ErrorType_General, _("Matrix size don't match") );

	if( width == 1 )
		return data[0];
	else if( width == 2 )
		return data[0] * data[3] - data[1] * data[2];
	else if( width == 3 )
		return data[0]*data[4]*data[8] + data[1]*data[5]*data[6] + data[2]*data[3]*data[7]
		         - ( data[0]*data[5]*data[7] + data[1]*data[3]*data[8] + data[2]*data[4]*data[6] );
	else
	{
		Matrix tmp;
		Complex det_res( Real(0) );
		Real sign(1);
		uint32 i,y,x,t;

		tmp.set_size(width-1,width-1);

		for(i=0; i < width ; i++,sign = -sign)
		{
			for(y=0; y < tmp.width ; y++)
			{
				for(x=0; x < tmp.width ; x++)
				{
					t = (y+1) * width + x;
					if(x >= i)
						t++;
					tmp.data[y * tmp.width + x] = data[t];
				}
			}
			det_res += sign * data[i] * tmp.det();
		}

		return det_res;
	}
}

bool Matrix::operator==( const Matrix &matr ) const
{
	uint32 x,y;
	if( width != matr.width  ||  height != matr.height )
		return false;
	else
	{
		for( y=1; y <= height; y++ )
		{
			for( x=1; x <= width; x++ )
			{
				if(get_complex(x,y) != matr.get_complex(x,y))
					return false;
			}
		}
	}
	return true;
}

void Matrix::set_next( const Complex &value ) throw(error_obj)
{
	if( next >= (width*height) )
		next = 0;
	if( (width*height) == 0 )
		return;
	data[next] = value;
	next++;
}
