
/*
 *  Matrix - Implements a matrix
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */


#ifndef MATRIX_H_
#define MATRIX_H_

#include "vartypes.h"
#include "integer.h"
#include "real.h"
#include "complex.h"
#include "utf8_string.h"



#define MATRIX_H_VERSION "v4"
#define MATRIX_H_DATE    "2005-10 - 2006-08"



class Matrix
{
private:
	uint32 width, height, next;

	Complex *data;

public:
	Matrix();
	Matrix(const Matrix &src) throw(error_obj);
	~Matrix();

	void operator=(const Matrix &src) throw(error_obj);

	bool operator==(const Matrix &matr) const;
	inline bool operator!=(const Matrix &matr) const { return !(this->operator==(matr)); }

	// sets the size of the Matrix
	void set_size(uint32 newwidth, uint32 newheight) throw(error_obj);

	inline uint32 get_width(void) const { return width; }
	inline uint32 get_height(void) const { return height; }

	// if col,row is outside the Matrix then nothing happens
	void set_complex(uint32 col,uint32 row,const Complex &value) throw(error_obj);

	// if width or height is zero nothing happens
	// data[next++] = value; if( next >= width*height ) next = 0; // this isn't exactly but almost as it is in the code
	void set_next( const Complex &value ) throw(error_obj);
	inline void set_next_index( uint32 new_next ) { next = new_next; }

	// if col,row is outside the Matrix then 0+i*0 is returned
	Complex get_complex(uint32 col,uint32 row) const throw(error_obj);

	Complex det(void) const throw(error_obj);

	void trans() throw(error_obj);

	void fill(const Complex &val) throw(error_obj);

	void identity(void) throw(error_obj);

	void adjoint(void) throw(error_obj);

	void inverse(void) throw(error_obj);

	// swaps the two rows with each other...if the Matrix doesn't have that many rows nothing happens
	void row_swap(uint32 row1,uint32 row2) throw(error_obj);

	// multiplies a row with val...if the Matrix doesn't have that many rows nothing happens
	void row_mul(uint32 row,const Complex &val) throw(error_obj);

	// temporary multiplies src_row with val and add it with dest_row...if any of src_row & dest_row
	// is too high for the Matrix nothing happens
	void row_mul_add(uint32 src_row,const Complex &val,uint32 dest_row) throw(error_obj);

	// this function replaces one row with one row from *src
	// if any of src_row & dest_row is too high for the Matrix's nothing happens
	void replace_row(uint32 dest_row, const Matrix *src, uint32 src_row) throw(error_obj);

	// calculates dest = src1 + src2
	friend void matrix_add(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj);

	// calculates dest = src1 - src2
	friend void matrix_sub(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj);

	// calculates this *= val
	void mul( const Complex &val ) throw(error_obj);

	// calculates dest = src1 * src2
	friend void matrix_mul(Matrix *dest,const Matrix *src1,const Matrix *src2) throw(error_obj);

	// calculates dest = src ^ power
	friend void matrix_pow( Matrix *dest, const Matrix *src, Integer power ) throw(error_obj);

	void append_to_string( utf8_string &str, const Format &fmt, bool row_break=false ) const
	       throw(error_obj);
};



#endif // MATRIX_H_
