
/*
 *  picture-object...
 *  Copyright (c) 2002-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See picture.h
 */

/*
v16  2006-07-28
---

	Replaced all throw_... function calls (as they no longer exists) with THROW_ERROR

v15  2006-04-22
---

	Picture's destructor is now virtual

v14  2005-09-15 - 2005-10-23
---

	Removed the support for delayed picture copying
	Added function draw_lines (which draws multiple lines with or without antialiasing)
	Rewrote filled_circle to use Bresenham's algorithm
	Added support for antialiasing in circle, filled_circle & ellipse
	Added operator==()
	Removed PictureType_YCrCb, PictureType_PALETTE as they wasn't supported anyway

v1.0.12  2005-09-09 - 2005-09-11
-------

	Removed class textpicture, as it wasn't used and I will build a vector-based replacer
	Rewrote circle to use Bresenham's algorithm 

v1.0.11  2005-08-08 - 2005-08-18
-------

	Replaced all asin(1.0) with M_PI similarities
	Added picture::line_antialiasing()

v1.0.10  2005-02-03 - 2005-02-04
-------

	Added support for background pictures (ie pictures without solidness-buffer)
	Added support for delayed picture copying

v1.0.9  2004-10-03 & 2004-12-19
------

	Removed unused variables
	Small fixes in circle, filed_circle & ellipse
	save_ppm now uses bool for save_as_ascii-argument
	Fixed some compiler warnings

v1.0.8  2004-07-14 - 2004-09-06
------

	Removed all appearance of Thrower
	Added copy constructor to picture
	Translation support completed
	Added picture::circle, picture::filled_circle, picture::ellipse
	Renamed a lot of functions
	Clear now also sets the solidness to 255 for the whole picture

v1.0.7  2004-07-03 - 2004-07-13
------

	Added some missed error checking
	Replaced some throw-lines with Throw_functions, to enable easy translation
	Fixed a complete check and recalculation of points outside picture in line-function
	Merged file picture.line.inc.cpp with picture.cpp
	Added color::ORANGE

v1.0.6  2004-06-17
------

	Replaced some const-groups with enum's, this changed the 'const'-names...

v1.0.5  2004-04-13 - 2004-04-17
------

	Added code in line-function so that lines that are totally outside the picture is ignored

v1.0.4  2003-12-18 - 2003-12-29 & 2004-03-20
------

	Small fix (2004-03-20)
	Pictures can be saved as PPM-files
	Added setpixel_grayscale (this function does nothing if the picture isn't grayscale)

v1.0.3  2003-02-22
------

	Added mirror, upsidedown, rotate90, rotate180 and rotate270
	The picture type grayscale is supported

v1.0.2  2003-02-04 - 2003-02-18
------

	textpicture now supports TEXT_RIGHT and TEXT_CENTER aligned text
	savepcx added to picture
	Added getwidth(),getheight() to textpicture

v1.0.1  2003-01-27
------

	Greatly reduced time usage for drawing pictures partly or totaly outside dest
	A speedup for RGB to RGB-picture-drawing
	A small speedup for drawshadow-function

v1.0.0  2002-12-14 - 2003-01-09
------

	Basic functions for the picture object
	PCX-files can be loaded
	Textpicture-object works
	The picture type RGB is supported
	The picture types YCrCb/grayscale/palette is not supported (but planned)
*/

#include "picture.h"
#include "keyfile.h"
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

namespace picture_h
{

picture *dest = 0;


bool colorrgb::operator==(const colorrgb &color) const
{
	if( red == color.red  &&  green == color.green  &&  blue == color.blue )
		return true;
	return false;
}

namespace color
{
	const colorrgb RED        = { 255,   0,   0 };
	const colorrgb GREEN      = {   0, 255,   0 };
	const colorrgb BLUE       = {   0,   0, 255 };
	const colorrgb DARK_RED   = { 170,   0,   0 };
	const colorrgb DARK_GREEN = {   0, 170,   0 };
	const colorrgb DARK_BLUE  = {   0,   0, 170 };
	const colorrgb YELLOW     = { 255, 255,   0 };
	const colorrgb WHITE      = { 255, 255, 255 };
	const colorrgb LIGHT_GRAY = { 170, 170, 170 };
	const colorrgb DARK_GRAY  = {  85,  85,  85 };
	const colorrgb BLACK      = {   0,   0,   0 };
	const colorrgb PURPLE     = { 165,   0, 255 };
	const colorrgb PINK       = { 255,   0, 229 };
	const colorrgb ORANGE     = { 255, 147,   0 };
}

picture::picture()
{
	width = 0;
	height = 0;
	type = PictureType_RGB;

	isBackground = false;

	data[0].buffer = 0;
	data[0].buffersize = 0;
	data[1].buffer = 0;
	data[1].buffersize = 0;
	data[2].buffer = 0;
	data[2].buffersize = 0;
	solidness.buffer = 0;
	solidness.buffersize = 0;
}
picture::picture(const picture &src) throw(error_obj)
{
	width = 0;
	height = 0;
	type = PictureType_RGB;

	isBackground = false;

	data[0].buffer = 0;
	data[0].buffersize = 0;
	data[1].buffer = 0;
	data[1].buffersize = 0;
	data[2].buffer = 0;
	data[2].buffersize = 0;
	solidness.buffer = 0;
	solidness.buffersize = 0;

	*this = src;
}

picture::~picture()
{
	erase();
}

void picture::set_isBackground(bool value) throw(error_obj)
{
	if( value )
	{
		delete [] solidness.buffer;
		solidness.buffer = 0;
		solidness.buffersize = 0;
		isBackground = value;
	}
	else
	{
		if( solidness.buffersize < (width*height) )
		{
			delete [] solidness.buffer;
			solidness.buffer = 0;
			solidness.buffersize = 0;
			try
			{
				solidness.buffer = new uint8 [ width*height ];
				solidness.buffersize = width*height;        
			}
			catch(...)
			{
				isBackground = true;
				THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
			}
		}
		isBackground = value;
	}
}


//
//  erase - freeing all the memory used by picture-object
//
void picture::erase(void)
{
	delete [] data[0].buffer;
	delete [] data[1].buffer;
	delete [] data[2].buffer;
	delete [] solidness.buffer;

	width = 0;
	height = 0;

	data[0].buffer = 0;
	data[0].buffersize = 0;
	data[1].buffer = 0;
	data[1].buffersize = 0;
	data[2].buffer = 0;
	data[2].buffersize = 0;
	solidness.buffer = 0;
	solidness.buffersize = 0;
}



//
//  setsize - changes the size of the image
//    Note: this is NOT a resizer, the imagedata will be destroyed
//
void picture::set_size(int32 newwidth,int32 newheight) throw(error_obj)
{
	int32 size,i;

	if( newwidth < 0  ||  newheight < 0 )
	{
		THROW_ERROR( ErrorType_General, _("Invalid width and/or height.") );
	}
	else if( newwidth == 0  ||  newheight == 0 )
	{
		erase();
		return;
	}
	else if( newwidth == width  &&  newheight == height )
		return;

	try
	{
		size = newwidth * newheight;

		if( !isBackground )
		{
			if(solidness.buffersize < size)
			{
				delete [] solidness.buffer;
				solidness.buffer = 0;
				solidness.buffersize = 0;

				solidness.buffer = new uint8 [size];

				solidness.buffersize = size;
			}
		}

		for(i=0;i<3;i++)
		{
			if( i != 0  &&  type == PictureType_GRAYSCALE )
				break;

			if(data[i].buffersize < size)
			{
				delete [] data[i].buffer;
				data[i].buffer = 0;
				data[i].buffersize = 0;

				data[i].buffer = new uint8 [size];

				data[i].buffersize = size;
			}
		}
		width = newwidth;
		height = newheight;
	}
	catch(...)
	{
		erase();
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
}

void picture::circle(int32 x, int32 y, uint32 r, const colorrgb &color, bool antialiased )
                throw(error_obj)
{
	if( antialiased )
	{
		uint32 nroflines,i, *lines=0;
		double angle,dr,step,dx,dy,dcx,dcy;
		float *vertex=0;

		dr = double(r);
		step = M_PI * dr / 2.0;

		if(step > 100.0)
			nroflines = 100;
		else
			nroflines = uint32(step + 0.5);
		if( nroflines < 6 )
			nroflines = 6;

		step = 2.0*M_PI / double(nroflines);
		angle = step;

		try
		{
			dcx = double(x);
			dcy = double(y);

			vertex = new float [nroflines*2+2];
			lines = new uint32 [nroflines*2];

			vertex[0] = float(dcx+dr);
			vertex[1] = float(dcy);

			for(i=0;i<nroflines;i++,angle+=step)
			{
				dx = cos(angle) * dr;
				dy = sin(angle) * dr;
	
				vertex[i*2+2] = float(dx+dcx);
				vertex[i*2+3] = float(dy+dcy);
				lines[i*2] = i*2;
				lines[i*2+1] = i*2+2;
			}
			draw_lines( vertex, lines, nroflines, color, true );
		}
		catch(...)
		{
			delete [] vertex;
			delete [] lines;
			THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
		}
		delete [] vertex;
		delete [] lines;
	}
	else
	{
// Using Bresenham's algorithm
		if( r != 0 )
		{
			int32 d, tx, ty;
		
			tx = 0;
			ty = int32(r);
			d = 3 - 2*ty;
	
			while(true)
			{
				set_pixel( x + tx, y + ty, color );
				set_pixel( x + tx, y - ty, color );
				set_pixel( x - tx, y + ty, color );
				set_pixel( x - tx, y - ty, color );
				set_pixel( x + ty, y + tx, color );
				set_pixel( x + ty, y - tx, color );
				set_pixel( x - ty, y + tx, color );
				set_pixel( x - ty, y - tx, color );
				
				if( tx >= ty )
					break;
	
				if( d < 0 )
					d += 4*tx + 6;
				else
				{
					d += 4*(tx - ty) + 10;
					ty--;
				}
				tx++;
			}
		}
	}
}

void picture::ellipse( int32 x, int32 y, uint32 rx, uint32 ry, const colorrgb &color,
                       bool antialiased ) throw(error_obj)
{
	uint32 nroflines,i, *lines=0;
	int32 nx,ny,prevx,prevy;
	double angle,drx,dry,step,dx,dy,dcx,dcy;
	float *vertex=0;
	
	drx = double(rx);
	dry = double(ry);
	step = M_PI * ( (drx > dry) ? drx : dry ) / 2.0;
	
	if(step > 100.0)
		nroflines = 100;
	else
		nroflines = uint32(step + 0.5);
	if( nroflines < 6 )
		nroflines = 6;

	step = 2.0*M_PI / double(nroflines);
	angle = step;

	try
	{
		if( antialiased )
		{
			dcx = double(x);
			dcy = double(y);

			vertex = new float [nroflines*2+2];
			lines = new uint32 [nroflines*2];

			vertex[0] = float(dcx+drx);
			vertex[1] = float(dcy);
		}
		else
		{
			prevx = x + rx;
			prevy = y;
		}

		for(i=0;i<nroflines;i++,angle+=step)
		{
			dx = cos(angle) * drx;
			dy = sin(angle) * dry;

			if( antialiased )
			{
				vertex[i*2+2] = float(dx+dcx);
				vertex[i*2+3] = float(dy+dcy);
				lines[i*2] = i*2;
				lines[i*2+1] = i*2+2;
			}
			else
			{
				nx = x + int32(dx + 0.5);
				ny = y + int32(dy + 0.5);
				line(prevx,prevy,nx,ny,color);

				prevx = nx;
				prevy = ny;
			}
		}
		if( antialiased )
			draw_lines( vertex, lines, nroflines, color, true );
	}
	catch(...)
	{
		delete [] vertex;
		delete [] lines;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
	delete [] vertex;
	delete [] lines;
}

void picture::filled_circle(int32 x, int32 y, uint32 r, const colorrgb &color, bool antialiased)
                throw(error_obj)
{
// Using Bresenham's algorithm
	if( r != 0 )
	{
		if( antialiased )
		{
			float cx, cy, fr = float(r), row, fx, fr2, t;
			cx = float(x) + 0.5f;
			cy = float(y) + 0.5f;
			fr2 = fr*fr;
			fx = fr;
			
			for( row=0.0f; row<fx; row++ )
			{
				fx = sqrt( fr2 - row*row );

				if( fx != trunc(fx) )
				{
					t = fx - trunc(fx);
					fx = trunc(fx);

					set_pixel( int32(cx-fx)-1, y+int32(row), color, uint8( 255.0*t) );
					set_pixel( int32(cx-fx)-1, y-int32(row), color, uint8( 255.0*t) );
					set_pixel( y+int32(row), int32(cx-fx)-1, color, uint8( 255.0*t) );
					set_pixel( y-int32(row), int32(cx-fx)-1, color, uint8( 255.0*t) );
					if( row != 0.0f )
					{
						set_pixel( int32(cx+fx)+1, y+int32(row), color, uint8( 255.0*t) );
						set_pixel( int32(cx+fx)+1, y-int32(row), color, uint8( 255.0*t) );
						set_pixel( y+int32(row), int32(cx+fx)+1, color, uint8( 255.0*t) );
						set_pixel( y-int32(row), int32(cx+fx)+1, color, uint8( 255.0*t) );
					}
				}
				
				line( int32(cx-fx), y+int32(row), int32(cx+fx), y+int32(row), color );
				line( int32(cx-fx), y-int32(row), int32(cx+fx), y-int32(row), color );
				line( y+int32(row), int32(cx-fx), y+int32(row), int32(cx+fx), color );
				line( y-int32(row), int32(cx-fx), y-int32(row), int32(cx+fx), color );
			}
		}
		else
		{
			int32 d, tx, ty, oldy;
	
			tx = 0;
			ty = int32(r);
			d = 3 - 2*ty;

			oldy = ty;
			while(true)
			{
				if( oldy != ty )
				{
					line( x - (tx-1), y + oldy, x + tx-1, y + oldy, color );
					line( x - (tx-1), y - oldy, x + tx-1, y - oldy, color );
					oldy = ty;
				}
				line( x - ty, y + tx, x + ty, y + tx, color );
				line( x - ty, y - tx, x + ty, y - tx, color );
			
				if( tx >= ty )
					break;

				if( d < 0 )
					d += 4*tx + 6;
				else
				{
					d += 4*(tx - ty) + 10;
					ty--;
				}
				tx++;
			}
		}
	}
}

//
//  line - draws a line on the picture
//
void picture::line_antialiasing(float x1,float y1,float x2,float y2,const colorrgb &color,
                                uint8 solid) throw(error_obj)
{
	float dx, dy, gradient, xend, yend, gap, tmp, fheight, fwidth;
	int32 xpxl1, xpxl2, ypxl1, ypxl2;

	if(y1 > y2) // makes sure that y1 is lower than y2
	{
		tmp = x1;
		x1 = x2;
		x2 = tmp;
		tmp = y1;
		y1 = y2;
		y2 = tmp;
	}
	fheight = float(height);
	fwidth = float(width);

// picture_line_calc_p1
	if(y1 >= fheight)
		return; // whole line is below the picture
	if(y1 < 0.0f)
	{
		if(y2 < 0.0f)
			return; // whole line is over the picture

		x1 = x1 - ( y1 * (x2 - x1) ) / (y2 - y1);
		y1 = 0.0f;
	}
	if(x1 < 0.0f)
	{
		if(x2 < 0.0f)
			return; // whole line is to the left of the picture

		y1 = y1 - ( x1 * (y2 - y1) ) / (x2 - x1);
		x1 = 0.0f;
		if(y1 >= fheight)
			return; // the line is to the left and below of the picture
	}
	else if(x1 >= fwidth)
	{
		if(x2 >= fwidth)
			return; // the line is to the right...

		y1 = y1 + ( (fwidth - 1.0f - x1) * (y2 - y1) ) / (x2 - x1);
		x1 = fwidth - 1.0f;
		if(y1 >= fheight)
			return; // the line is to the right and below...
	}


// picture_line_calc_p2
	if(y2 >= fheight)
	{
		x2 = x1 + ( (fheight - 1.0f - y1) * (x2 - x1) ) / (y2 - y1);
		y2 = fheight - 1.0f;
	}
	if(x2 < 0.0f)
	{
		y2 = y1 - ( x1 * (y2 - y1) ) / (x2 - x1);
		x2 = 0.0f;
	}
	else if(x2 >= fwidth)
	{
		y2 = y1 + ( (fwidth - 1.0f - x1) * (y2 - y1) ) / (x2 - x1);
		x2 = fwidth - 1.0f;
	}

	dx = x2 - x1;
	dy = y2 - y1;

	if( fabs(dx) > fabs(dy) )
	{
		if( x2 < x1 )
		{
			tmp = x1;
			x1 = x2;
			x2 = tmp;
			tmp = y1;
			y1 = y2;
			y2 = tmp;
		}
		gradient = dy / dx;

// handle first endpoint
		xend = round(x1);
		yend = y1 + gradient * (xend - x1);
		gap = 1.0 - (x1+0.5 - trunc(x1 + 0.5));
		xpxl1 = int32(xend);  // this will be used in the main loop
		ypxl1 = int32( trunc(yend) );
		set_pixel( xpxl1, ypxl1, color, uint8( float(solid)* (1.0 - (yend - trunc(yend)) * gap) ) );
		set_pixel( xpxl1, ypxl1 + 1, color, uint8( float(solid)* (yend - trunc(yend)) * gap ) );
		tmp = yend + gradient; // first y-intersection for the main loop

// handle second endpoint
		xend = round(x2);
		yend = y2 + gradient * (xend - x2);
		gap = 1.0 - (x2-0.5 - trunc(x2 - 0.5));
		xpxl2 = int32(xend);  // this will be used in the main loop
		ypxl2 = int32( trunc(yend) );
		set_pixel( xpxl2, ypxl2, color, uint8( float(solid)* (1.0 - (yend - trunc(yend)) *gap) ) );
		set_pixel( xpxl2, ypxl2 + 1, color, uint8( float(solid)* (yend - trunc(yend)) *gap ) );

// main loop
		for( xpxl1++; xpxl1< xpxl2; xpxl1++ )
		{
			set_pixel( xpxl1, int32(trunc(tmp)), color, uint8( float(solid)* (1.0 - (tmp - trunc(tmp))) ) );
			set_pixel( xpxl1, int32(trunc(tmp)) + 1, color, uint8( float(solid)* (tmp - trunc(tmp)) ) );
			tmp += gradient;
		}
	}
	else
	{
		if( y2 < y1 )
		{
			tmp = x1;
			x1 = x2;
			x2 = tmp;
			tmp = y1;
			y1 = y2;
			y2 = tmp;
		}
		gradient = dx / dy;

// handle first endpoint
		yend = round(y1);
		xend = x1 + gradient * (yend - y1);
		gap = 1.0 - (y1+0.5 - trunc(y1 + 0.5));
		ypxl1 = int32(yend);  // this will be used in the main loop
		xpxl1 = int32( trunc(xend) );
		set_pixel( xpxl1, ypxl1, color, uint8( float(solid)* (1.0 - (xend - trunc(xend)) * gap) ) );
		set_pixel( xpxl1 + 1, ypxl1, color, uint8( float(solid)* (xend - trunc(xend)) * gap ) );
		tmp = xend + gradient; // first y-intersection for the main loop

// handle second endpoint
		yend = round(y2);
		xend = x2 + gradient * (yend - y2);
		gap = 1.0 - (y2-0.5 - trunc(y2 - 0.5));
		ypxl2 = int32(yend);  // this will be used in the main loop
		xpxl2 = int32( trunc(xend) );
		set_pixel( xpxl2, ypxl2, color, uint8( float(solid)* (1.0 - (xend - trunc(xend)) *gap) ) );
		set_pixel( xpxl2 + 1, ypxl2, color, uint8( float(solid)* (xend - trunc(xend)) *gap ) );

// main loop
		for( ypxl1++; ypxl1< ypxl2; ypxl1++ )
		{
			set_pixel( int32(trunc(tmp)), ypxl1, color, uint8( float(solid)* (1.0 - (tmp - trunc(tmp))) ) );
			set_pixel( int32(trunc(tmp)) + 1, ypxl1, color, uint8( float(solid)* (tmp - trunc(tmp)) ) );
			tmp += gradient;
		}
	}
}



//
//  line - draws a line on the picture
//
void picture::line(int32 x1,int32 y1,int32 x2,int32 y2,const colorrgb &color,
                   uint8 solid) throw(error_obj)
{
	int32 dx,dy,incr1,incr2,d,x,y,xend,yend,yinc,xinc;

	float fx1,fx2,fy1,fy2,fheight,fwidth; // the calculations are done using float's to avoid
// overflows on very large values

	if(y1 == y2)
		goto picture_line_draw_h_line;

	if(y1 > y2) // makes sure that y1 is lower than y2
	{
		xend = x1;
		x1 = x2;
		x2 = xend;
		xend = y1;
		y1 = y2;
		y2 = xend;
	}
	fx1 = float(x1);
	fx2 = float(x2);
	fy1 = float(y1);
	fy2 = float(y2);
	fheight = float(height);
	fwidth = float(width);

// picture_line_calc_p1
	if(y1 >= height)
		return; // whole line is below the picture
	if(y1 < 0)
	{
		if(y2 < 0)
			return; // whole line is over the picture

		fx1 = fx1 - ( fy1 * (fx2 - fx1) ) / (fy2 - fy1);
		fy1 = 0.0f;
	}
	if(fx1 < 0.0f)
	{
		if(fx2 < 0.0f)
			return; // whole line is to the left of the picture

		fy1 = fy1 - ( fx1 * (fy2 - fy1) ) / (fx2 - fx1);
		fx1 = 0.0f;
		if(fy1 >= fheight)
			return; // the line is to the left and below of the picture
	}
	else if(fx1 >= fwidth)
	{
		if(fx2 >= fwidth)
			return; // the line is to the right...

		fy1 = fy1 + ( (fwidth - 1.0f - fx1) * (fy2 - fy1) ) / (fx2 - fx1);
		fx1 = fwidth - 1.0f;
		if(fy1 >= fheight)
			return; // the line is to the right and below...
	}


// picture_line_calc_p2
	if( trunc(fy1) != trunc(fy2) )
	{
		if( fy2 >= fheight )
		{
			fx2 = fx1 + ( (fheight - 1.0f - fy1) * (fx2 - fx1) ) / (fy2 - fy1);
			fy2 = fheight - 1.0f;
		}
		if( fx2 < 0.0f )
		{
			fy2 = fy1 - ( fx1 * (fy2 - fy1) ) / (fx2 - fx1);
			fx2 = 0.0f;
		}
		else if( fx2 >= fwidth )
		{
			fy2 = fy1 + ( (fwidth - 1.0f - fx1) * (fy2 - fy1) ) / (fx2 - fx1);
			fx2 = fwidth - 1.0f;
		}

		x1 = int32(trunc(fx1));
		x2 = int32(trunc(fx2));
		y1 = int32(trunc(fy1));
		y2 = int32(trunc(fy2));
	}
	else
	{
		x1 = int32(trunc(fx1));
		x2 = int32(trunc(fx2));
		y1 = int32(trunc(fy1));

		picture_line_draw_h_line:
		if( y1 >= height  ||  y1 < 0 )
			return; // the whole line is over/below the picture
		if( x1 > x2 )
		{
			xend=x1;
			x1=x2;
			x2=xend;
		}

		if( x1 >= width  ||  x2 < 0 )
			return; // the whole line is to the right/left of the picture

		if( x1 < 0 )
			x1 = 0;
		if( x2 >= width )
			x2 = width-1;

		if( x1 > x2 )
			return; // the whole line is to the right/left of the picture

		if( solid == 255  &&  type == PictureType_RGB )
		{
			xend = x1 + y1*width; // offset
			yend = x2 - x1 + 1;   // length

			memset( &data[0].buffer[xend], color.red, yend);
			memset( &data[1].buffer[xend], color.green, yend);
			memset( &data[2].buffer[xend], color.blue, yend);

			if( !isBackground )
				memset( &solidness.buffer[xend], 255, yend );
		}
		else
		{
			for(;x1 <= x2;x1++)
				set_pixel(x1,y1,color,solid);
		}
		return;
	}

	set_pixel(x1,y1,color,solid);
	set_pixel(x2,y2,color,solid);

	dx=abs(x2-x1);
	dy=abs(y2-y1);
	if(dx>=dy)
	{
		if(x1>x2)
		{
			x=x2;
			y=y2;
			xend=x1;
			if(dy==0)
				yinc=0;
			else
			{
				if(y2>y1)
					yinc=-1;
				else
					yinc=1;
			}
		}
		else
		{
			x=x1;
			y=y1;
			xend=x2;
			if(dy==0)
				yinc=0;
			else
			{
				if(y2>y1)
					yinc=1;
				else
					yinc=-1;
			}
		}
		incr1=dy+dy;
		d=incr1-dx;
		incr2=(dy-dx)<<1;
		while(x<xend)
		{
			x++;
			if(d<0)
				d+=incr1;
			else
			{
				y+=yinc;
				d+=incr2;
			}
			set_pixel(x,y,color,solid);
		}
	}
	else
	{
		if(y1>y2)
		{
			x=x2;
			y=y2;
			yend=y1;
			if(dx==0)
				xinc=0;
			else
			{
				if(x2>x1)
					xinc=-1;
				else
					xinc=1;
			}
		}
		else
		{
			x=x1;
			y=y1;
			yend=y2;
			if( dx == 0 )
				xinc=0;
			else
			{
				if(x2>x1)
					xinc=1;
				else
					xinc=-1;
			}
		}
		incr1=dx+dx;
		d=incr1-dy;
		incr2=(dx-dy)<<1;
		while( y < yend )
		{
			y++;
			if( d < 0 )
				d+=incr1;
			else
			{
				x+=xinc;
				d+=incr2;
			}
			set_pixel(x,y,color,solid);
		}
	}
}


//
//  setsolidness - sets the solidness of a specific pixel
//
void picture::set_solidness(int32 x,int32 y,uint8 newsolidness) throw(error_obj)
{
	int32 offset = x + y*width;

	if( !isBackground )
	{
		if((x < width) & (y < height) & (x >= 0) & (y >= 0))
			solidness.buffer[offset] = newsolidness;
	}
}


//
//  setsolidnesscolor - sets the solidness to one specific color
//
void picture::set_solidness_color(const colorrgb &color,uint8 newsolidness) throw(error_obj)
{
	int32 x,y;

	if( !isBackground )
	{
		for(y=0;y<height;y++)
		{
			for(x=0;x<width;x++)
			{
				if(color == get_pixel(x,y))
					set_solidness(x,y,newsolidness);
			}
		}
	}
}


//
//  setsolidnessall - sets the solidness of the whole picture
//
void picture::set_solidness_all(uint8 newsolidness) throw(error_obj)
{
	if( !isBackground )
		memset(solidness.buffer,newsolidness, height * width);
}


//
//  clear - set all pixels to the specified color
//
void picture::clear(const colorrgb &color) throw(error_obj)
{
	int32 size = height*width;

	if( !isBackground )
		memset(solidness.buffer, 255, size );
	if( type == PictureType_RGB )
	{
		memset(data[0].buffer, color.red,   size);
		memset(data[1].buffer, color.green, size);
		memset(data[2].buffer, color.blue,  size);
	}
	else //if( type == PictureType_GRAYSCALE )
	{
		uint8 grey = uint8(
		                   float(color.red)*0.299f +
		                   float(color.green)*0.587f +
		                   float(color.blue)*0.114f
		                  );
		memset(data[0].buffer, grey, size);
	}
}


//
//  convto_rgb - converts the picture into a rgb-picture
//
void picture::convto_rgb(void) throw(error_obj)
{
	if( type == PictureType_GRAYSCALE )
	{
		set_size(width,height); // fixing space for the green and blue channel
		type = PictureType_RGB;
		memcpy(data[1].buffer,data[0].buffer,width*height);
		memcpy(data[2].buffer,data[0].buffer,width*height);
	}
//	else if( type == PictureType_RGB ) // this is already a rgb-picture
//	{ }
}

//
//  convto_grayscale - converts the picture into a grayscaled-picture
//
void picture::convto_grayscale(void) throw(error_obj)
{
//Y = 0.299*R + 0.587*G + 0.114*B
	int32 offset = width * height -1;

	if( type == PictureType_RGB )
	{
		for(;offset >= 0;offset--)
			data[0].buffer[offset] = uint8(
			                               float(data[0].buffer[offset])*0.299f +
			                               float(data[1].buffer[offset])*0.587f +
			                               float(data[2].buffer[offset])*0.114f
			                              );
		type = PictureType_GRAYSCALE;
	}
//	else if( type == PictureType_GRAYSCALE ) // this is already a grayscale-picture
//	{ }
}


void picture::set_pixel_grayscale(int32 x,int32 y,uint8 grey,uint8 solid) throw(error_obj)
{
	if((x < width) & (y < height) & (x > 0) & (y > 0) & (type == PictureType_GRAYSCALE))
	{
		if(solid == 255)
			data[0].buffer[x+y*width] = grey;
		else
			data[0].buffer[x+y*width] = uint8( (uint16(grey) * uint16(solid) + uint16(data[0].buffer[x+y*width]) * uint16(255 - solid)) / 255);

		if( solid != 0  &&  !isBackground )
			solidness.buffer[x+y*width] = 255;
	}
}

//
//  setpixel - sets a pixel to a specific color
//
//  the color will be converted to grayscale/YCrCb/palette if the picture is in those modes
//
//  if the point (x,y) is outside the picture nothing will happen
//
//  the picture-pixel's solidness will be 255
//
void picture::set_pixel(int32 x,int32 y,const colorrgb &color,uint8 solid) throw(error_obj)
{
	int32 offset = x + y*width;

	if((x < width) & (y < height) & (x >= 0) & (y >= 0))
	{
		if( type == PictureType_RGB )
		{
			if(solid == 255)
			{
				data[0].buffer[offset] = color.red;
				data[1].buffer[offset] = color.green;
				data[2].buffer[offset] = color.blue;
			}
			else
			{
				data[0].buffer[offset] = uint8( (uint16(color.red) * uint16(solid) + uint16(data[0].buffer[offset]) * uint16(255 - solid)) / 255);
				data[1].buffer[offset] = uint8( (uint16(color.green) * uint16(solid) + uint16(data[1].buffer[offset]) * uint16(255 - solid)) / 255);
				data[2].buffer[offset] = uint8( (uint16(color.blue) * uint16(solid) + uint16(data[2].buffer[offset]) * uint16(255 - solid)) / 255);
			}
			if( solid != 0  &&  !isBackground)
				solidness.buffer[offset] = 255;
		}
		else //if( type == PictureType_GRAYSCALE )
		{
			uint8 gray;

			if((color.red == color.green) & (color.red == color.green))
				gray = color.red;
			else
			{
				gray = uint8(
							 float(color.red)*0.299f +
							 float(color.green)*0.587f +
							 float(color.blue)*0.114f
							);
			}
			if(solid == 255)
				data[0].buffer[offset] = gray;
			else
				data[0].buffer[offset] = uint8( (uint16(gray) * uint16(solid) + uint16(data[0].buffer[offset]) * uint16(255 - solid)) / 255);
			if( solid != 0  &&  !isBackground )
				solidness.buffer[offset] = 255;

		}
	}
}


void picture::box(int32 x,int32 y,int32 width,int32 height,const colorrgb &color,
                  uint8 solid) throw(error_obj)
{
	if( width <= 0  ||  height <= 0 )
		return;
	width--;
	height--;
	line(x,y,x+width,y,color,solid);
	line(x,y+height,x+width,y+height,color,solid);
	line(x,y,x,y+height,color,solid);
	line(x+width,y,x+width,y+height,color,solid);
}

void picture::filled_box(int32 x,int32 y,int32 width,int32 height,const colorrgb &color,
                         uint8 solid) throw(error_obj)
{
	int32 ystop,xstop;

	if( width <= 0  ||  height <= 0 )
		return;

	xstop = x+width-1;
	ystop = y+height-1;

	for(;y<=ystop;y++)
		line(x,y,xstop,y,color,solid);
}



//
//  draw - draws this picture on *dest
//
void picture::draw(int32 x,int32 y) const throw(error_obj)
{
	int32 xt,yt,xstart,ystart,xstop,ystop;
	colorrgb color;
	uint8 solid;

	if(dest == 0)
		return;

	if(x < 0)
		xstart = -x;
	else
		xstart = 0;

	if(y < 0)
		ystart = -y;
	else
		ystart = 0;

	if(width > (dest->width - x))
		xstop = dest->width - x;
	else
		xstop = width;

	if(height > (dest->height - y))
		ystop = dest->height - y;
	else
		ystop = height;

	if( type == PictureType_RGB  &&  dest->type == PictureType_RGB )
	{
		int32 offset = xstart + ystart*width;
		int32 lineskip = xstart + width - xstop;

		for(yt=ystart;yt<ystop;yt++,offset += lineskip)
		{
			for(xt=xstart;xt<xstop;xt++,offset++)
			{
				dest->set_pixel(xt+x,yt+y,(colorrgb) {data[0].buffer[offset],
				data[1].buffer[offset],
				data[2].buffer[offset]},
				solidness.buffer[offset]);
			}
		}
	}
	else
	{
		for(yt=ystart;yt<ystop;yt++)
		{
			for(xt=xstart;xt<xstop;xt++)
			{
				color = get_pixel(xt,yt);
				solid = get_solidness(xt,yt);
				dest->set_pixel(xt+x,yt+y,color,solid);
			}
		}
	}
}


//
//  draw_shadow - draws a shadow of the picture in the specified color
//
void picture::draw_shadow(int32 x,int32 y,const colorrgb &color) const throw(error_obj)
{
	int32 xt,yt,xstart,ystart,xstop,ystop;

	if(dest == 0)
		return;

	if(x < 0)
		xstart = -x;
	else
		xstart = 0;

	if(y < 0)
		ystart = -y;
	else
		ystart = 0;

	if(width > (dest->width - x))
		xstop = dest->width - x;
	else
		xstop = width;

	if(height > (dest->height - y))
		ystop = dest->height - y;
	else
		ystop = height;

	int32 offset = xstart + ystart*width;
	int32 lineskip = xstart + width - xstop;

	for(yt=ystart;yt<ystop;yt++, offset += lineskip)
	{
		for(xt=xstart;xt<xstop;xt++, offset++)
			dest->set_pixel(xt+x,yt+y,color,solidness.buffer[offset]);
	}
}

//
//  get_pixel - returns a colorrgb-struct containing the rgb-values for the pixel
//              if picture isn't a RGB-picture the rgb-value will be calculated
//
colorrgb picture::get_pixel(int32 x,int32 y) const
{
	if( x < 0  ||  y < 0  ||  x >= width  ||  y >= height )
		return color::BLACK;

	int32 offset = x+y*width;
	if( type == PictureType_RGB )
		return (colorrgb) { data[0].buffer[offset], data[1].buffer[offset], data[2].buffer[offset] };
	else //if( type == PictureType_GRAYSCALE )
		return (colorrgb) { data[0].buffer[offset], data[0].buffer[offset], data[0].buffer[offset] };
}



uint8 picture::get_solidness(int32 x,int32 y) const
{
	if( x < 0  ||  y < 0  ||  x >= width  ||  y >= height )
		return 0;

	if( isBackground )
		return 255;
	else
		return solidness.buffer[x+y*width];
}


void picture::save_ppm(const char *filename, bool save_as_ascii) const throw(error_obj)
{
	FILE *fp;
	colorrgb farg;
	int32 x,y;

	fp = fopen(filename,"wb");
	if(fp == NULL)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/write to file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	if(save_as_ascii)
		fprintf(fp,"P3\n");
	else
		fprintf(fp,"P6\n");

	fprintf(fp,"%ld\n%ld\n255%c",width,height,10);

	for(y=0;y<height;y++)
	{
		for(x=0;x<width;x++)
		{
			farg = get_pixel(x,y);

			if(save_as_ascii)
				fprintf(fp,"%lu %lu %lu\n",uint32(farg.red),uint32(farg.green),uint32(farg.blue));
			else
			{
				fputc(farg.red,fp);
				fputc(farg.green,fp);
				fputc(farg.blue,fp);
			}
		}
	}

	if(fclose(fp) != 0)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't close file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}
}


void picture::save_pcx(const char *filename) const throw(error_obj)
{
	FILE *fp;
	uint8 header[128],tmp,lastbyte;
	int32 x,y,i,antal;
	colorrgb farg;

	fp = fopen(filename,"wb");
	if(fp == NULL)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/write to file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	memset(header,0,128);

	header[0] = 10; //ZSoft .PCX
	header[1] = 5;  //version 3.0
	header[2] = 1;  //RLE encoding
	header[3] = 8;  //bit depth
	header[65] = 3; //number of planes

	header[4] = 0;  //minX
	header[5] = 0;

	header[6] = 0;  //minY
	header[7] = 0;

	header[8] = uint8((width - 1) % 256); //maxX
	header[9] = uint8((width - 1) / 256);

	header[10] = uint8((height - 1) % 256);//maxY
	header[11] = uint8((height - 1) / 256);

	header[66] = uint8(width % 256); //bytes per line per plane
	header[67] = uint8(width / 256);

	for(i=0;i<128;i++)
		fputc(header[i],fp);

// saves picture
	for(y=0;y<height;y++)
	{
		for(i=0;i<3;i++)
		{
			antal = 1;
			farg = get_pixel(0,y);
			if(i == 0)
				lastbyte = farg.red;
			else if(i == 1)
				lastbyte = farg.green;
			else
				lastbyte = farg.blue;


			for(x=1;x<width;x++)
			{
				farg = get_pixel(x,y);
				if(i == 0)
					tmp = farg.red;
				else if(i == 1)
					tmp = farg.green;
				else
					tmp = farg.blue;

				if(tmp == lastbyte)
					antal++;
				else
				{
saveit_pcx_picture:
					if((antal != 1) | (lastbyte >= uint8(0xC0)))
					{
						tmp = uint8(antal);
						if(antal > int32(0x3F))
							tmp = uint8(0x3F);
						antal -= int32(tmp);

						tmp += uint8(0xC0);
						fputc(tmp,fp);
						fputc(lastbyte,fp);
						if(antal > 0)
							goto saveit_pcx_picture;
					}
					else
						fputc(lastbyte,fp);

					antal = 1;
					farg = get_pixel(x,y);
					if(i == 0)
						lastbyte = farg.red;
					else if(i == 1)
						lastbyte = farg.green;
					else
						lastbyte = farg.blue;
				}
				if(x == (width - 1))
				{
					x = width;
					goto saveit_pcx_picture;
				}
			}
		}
	}
	if(fclose(fp) != 0)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't close file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}
}


void picture::mirror(void) throw(error_obj)
{
	colorrgb color;
	uint8 solid;
	int32 x,y;

	for(y=0;y<height;y++)
	{
		for(x=0;x<(width/2);x++)
		{
			color = get_pixel(x,y);
			solid = get_solidness(x,y);
			set_pixel(x,y,get_pixel(width-1-x,y));
			set_solidness(x,y,get_solidness(width-1-x,y));
			set_pixel(width-1-x,y,color);
			set_solidness(width-1-x,y,solid);
		}
	}
}
void picture::upside_down(void) throw(error_obj)
{
	colorrgb color;
	uint8 solid;
	int32 x,y;

	for(y=0;y<(height/2);y++)
	{
		for(x=0;x<width;x++)
		{
			color = get_pixel(x,y);
			solid = get_solidness(x,y);
			set_pixel(x,y,get_pixel(x,height-1-y));
			set_solidness(x,y,get_solidness(x,height-1-y));
			set_pixel(x,height-1-y,color);
			set_solidness(x,height-1-y,solid);
		}
	}
}
void picture::rotate90(void) throw(error_obj)
{
	picture tmp;
	int32 x,y;

	tmp = *this;

	height = tmp.width;
	width = tmp.height;

	for(y=0;y<height;y++)
	{
		for(x=0;x<width;x++)
		{
			set_pixel(x,y,tmp.get_pixel(height-1-y,x));
			set_solidness(x,y,tmp.get_solidness(height-1-y,x));
		}
	}
}
void picture::rotate180(void) throw(error_obj)
{
	mirror();
	upside_down();
}
void picture::rotate270(void) throw(error_obj)
{
	rotate90();
	rotate180();
}

//
//  load_pcx - loads a PCX-picture
//  note that the PCX-file must be a 24bit RLE v3.0
//
void picture::load_pcx(const char *filename) throw(error_obj)
{
	uint8 header[128],byte,color;
	FILE *fp;
	int32 i,tmp,x,y;

	erase();
	type = PictureType_RGB;

	fp = fopen(filename,"rb"); // open pcx-file
	if(fp == NULL)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/read from file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	for(i=0;i<128;i++)
	{
		header[i] = fgetc(fp);  // load pcx-header
		if(feof(fp) != 0)
		{
			fclose(fp);
			char tmp_err[ERROR_OBJ_MSG_LEN];
			snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/read from file (%s)"), filename );
			THROW_ERROR( ErrorType_File_IO, tmp_err );
		}
	}

	if( header[0] != 10 ||  // 10 = ZSoft .PCX
	    header[1] != 5 ||   // version should be 5 = v3.0
	    header[2] != 1 ||   // encoding 1 = .PCX Run Length Encoding (RLE)
	    header[65] != 3 ||  // number of planes should be 3 (r,g,b)
	    header[3] != 8 )    // bits per pixel per plane should be 8 (8 bit)
	{
		fclose(fp);
	
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("%s is not a valid PCX file"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}

	tmp = int32(header[4]) + (int32(header[5])<<8); //getting the width
	x = int32(header[8]) + (int32(header[9])<<8);
	x -= (tmp - 1);


	tmp = int32(header[6]) + (int32(header[7])<<8); //getting the height
	y = int32(header[10]) + (int32(header[11])<<8);
	y -= (tmp - 1);


	try{ set_size(x,y); }
	catch(error_obj error)
	{
		fclose(fp);
		throw error;
	}
	if( !isBackground )
		memset(solidness.buffer,255,width*height);

	tmp = 0;
	for(y=0;y<height;y++)
	{
		for(i=0;i<3;i++)
		{
			x = 0;
			do{
				byte = fgetc(fp);
				if(feof(fp) != 0)
				{
					fclose(fp);

					char tmp_err[ERROR_OBJ_MSG_LEN];
					snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/read from file (%s)"), filename );
					THROW_ERROR( ErrorType_File_IO, tmp_err );
				}
				if(byte >= uint8(0xc0))
				{
					color = fgetc(fp);
					if(feof(fp) != 0)
					{
						fclose(fp);

						char tmp_err[ERROR_OBJ_MSG_LEN];
						snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't open/read from file (%s)"), filename );
						THROW_ERROR( ErrorType_File_IO, tmp_err );
					}

					for(byte -= uint8(0xc0);byte!=0;byte--)
					{
						if(x == width)
						{
							tmp = 1;
							break;
						}

						data[i].buffer[x + y*width] = color;

						x++;
					}
				}
				else
				{
					data[i].buffer[x + y*width] = byte;

					x++;
				}
			}while(x < width);
		}
	}

	if(fclose(fp) != 0)
	{
		char tmp_err[ERROR_OBJ_MSG_LEN];
		snprintf( tmp_err, ERROR_OBJ_MSG_LEN, _("Couldn't close file (%s)"), filename );
		THROW_ERROR( ErrorType_File_IO, tmp_err );
	}
}


void picture::operator=(const picture & source) throw(error_obj)
{
	int32 size = source.width*source.height;

	type = source.type;
	set_size(source.width,source.height);

	if( !isBackground )
		memcpy(solidness.buffer,source.solidness.buffer,size);
	memcpy(data[0].buffer,source.data[0].buffer,size);
		
	if( type == PictureType_RGB )
	{
		memcpy(data[1].buffer,source.data[1].buffer,size);
		memcpy(data[2].buffer,source.data[2].buffer,size);
	}
}

bool picture::operator==(const picture &pic) const
{
	if( width != pic.width  ||  height != pic.height )
		return false;
	if( type != pic.type  ||  isBackground != pic.isBackground )
		return false;

	int32 tot = width*height;
	int32 i;

	if( !isBackground )
	{
		for( i=0; i<tot; i++ )
		{
			if( solidness.buffer[i] != pic.solidness.buffer[i] )
				return false;
		}
	}
	for( i=0; i<tot; i++ )
	{
		if( data[0].buffer[i] != pic.data[0].buffer[i] )
			return false;
	}

	if( type == PictureType_RGB )
	{		
		for( i=0; i<tot; i++ )
		{
			if( data[1].buffer[i] != pic.data[1].buffer[i] )
				return false;
			if( data[2].buffer[i] != pic.data[2].buffer[i] )
				return false;
		}
	}
	return true;
}

//
//  capture - extracts a part of a picture
//
void picture::capture(const picture &source,int32 x,int32 y,int32 width,int32 height)
                throw(error_obj)
{
	int32 xi,yi,tx,x2,y2;

	if( width <= 0  ||  height <= 0 )
		return;

	x2 = x+width-1;
	y2 = y+height-1;

	type = PictureType_RGB;
	set_isBackground( source.isBackground );
	set_size( width, height );

	tx = x;
	for(yi=0;y<=y2;y++,yi++)
	{
		for(x=tx,xi=0;x<=x2;x++,xi++)
		{
			set_pixel(xi,yi,source.get_pixel(x,y));
			if( !isBackground )
				set_solidness(xi,yi,source.get_solidness(x,y));
		}
	}
}


void picture::draw_lines( const float *vertex, const uint32 *lines, uint32 nroflines,
                          const colorrgb &color, bool antialiased, uint8 solid ) throw(error_obj)
{
	if( nroflines == 0 )
		return;

	if( antialiased )
	{
		float x, y, min_dist, xmax, xmin, ymax, ymin;
		float x1, x2, y1, y2, tmp;
		uint32 i, i2;

		nroflines *= 2;

		for( i=0; i<nroflines; i+=2 )
		{
			xmin = vertex[lines[i]];
			ymin = vertex[lines[i]+1];
			x2 = vertex[lines[i+1]];
			y2 = vertex[lines[i+1]+1];

			if( xmin < -1.0f  &&  x2 < -1.0f ) // line to the left of picture
				continue;
			if( ymin < -1.0f  &&  y2 < -1.0f ) // line above picture
				continue;
			if( xmin > float(width+1)  &&  x2 > float(width+1) ) // line to the right of picture
				continue;
			if( ymin > float(height+1)  &&  y2 > float(height+1) ) // line below picture
				continue;

			xmax = xmin;
			ymax = ymin;
			xmin = (xmin<x2) ? xmin : x2;
			ymin = (ymin<y2) ? ymin : y2;
			xmax = (xmax>x2) ? xmax : x2;
			ymax = (ymax>y2) ? ymax : y2;

			xmin = trunc(xmin);
			ymin = trunc(ymin);
			xmax = trunc(xmax+1.0f);
			ymax = trunc(ymax+1.0f);

			xmin = (xmin<0.0f) ? 0.0f : xmin;
			ymin = (ymin<0.0f) ? 0.0f : ymin;
			xmax = (xmax>float(width-1)) ? float(width-1) : xmax;
			ymax = (ymax>float(height-1)) ? float(height-1) : ymax;

			for( y=ymin; y<=ymax; y++ )
			{
				for( x=xmin; x<=xmax; x++ )
				{
					for( i2=i-2; i2<nroflines; i2-=2 )
					{
						x1 = vertex[lines[i2]];
						y1 = vertex[lines[i2]+1];
						x2 = vertex[lines[i2+1]];
						y2 = vertex[lines[i2+1]+1];

						if( x1 > x2 )
						{
							tmp = x1;
							x1 = x2;
							x2 = tmp;
						}
						if( y1  > y2 )
						{
							tmp = y1;
							y1 = y2;
							y2 = tmp;
						}
						x1 = trunc(x1);
						y1 = trunc(y1);
						x2 = trunc(x2+1.0f);
						y2 = trunc(y2+1.0f);
						if( x >= x1  &&  x <= x2  &&  y >= y1  &&  y <= y2 )
							break;
					}
					if( i2 < nroflines )
						continue;

					for( i2=i, min_dist=1.0f; i2<nroflines; i2+=2 )
					{
						x1 = vertex[lines[i2]];
						y1 = vertex[lines[i2]+1];
						x2 = vertex[lines[i2+1]];
						y2 = vertex[lines[i2+1]+1];

						if( x < (( (x1<x2)?x1:x2 )-1.0f) )
							continue;
						if( x > (( (x1>x2)?x1:x2 )+1.0f) )
							continue;
						if( y < (( (y1<y2)?y1:y2)-1.0f) )
							continue;
						if( y > (( (y1>y2)?y1:y2 )+1.0f) )
							continue;

						if( x1 == x2 )
						{
							if( y1 == y2 )  // A point
								tmp = -1.0f;
							else  //  A vertical line
								tmp = (y-y1) / (y2-y1);
						}
						else
						{  //  An ordinary line
							tmp = (y1-y2)/(x2-x1);
							tmp = ( x1 - x - tmp*(y1-y) ) / ( tmp*(y2-y1) - x2 + x1 );
						}
						if( tmp < 0.0f )
						{  //  Closest point lies below the line
							tmp = (x-x1)*(x-x1) + (y-y1)*(y-y1);
							if( tmp < min_dist )
								min_dist = tmp;
						}
						else if( tmp > 1.0f )
						{  //  Closest point lies above the line
							tmp = (x-x2)*(x-x2) + (y-y2)*(y-y2);
							if( tmp < min_dist )
								min_dist = tmp;
						}
						else
						{  //  Closest point lies on the line
							x1 += tmp*(x2-x1); // Calculating the point which lies closest to (x,y)
							y1 += tmp*(y2-y1);

							tmp = (x-x1)*(x-x1) + (y-y1)*(y-y1);
							if( tmp < min_dist )
								min_dist = tmp;
						}
					}
					if( min_dist < 1.0f )
						set_pixel( int32(x), int32(y), color,
						           uint8( float(solid)* (1.0f-sqrt(min_dist)) ) );
				}
			}
		}
	}
	else
	{
		nroflines *= 2;
		for( uint32 i=0; i<nroflines; i+=2 )
			line( int32(vertex[lines[i]]), int32(vertex[lines[i]+1]),
			      int32(vertex[lines[i+1]]), int32(vertex[lines[i+1]+1]), color, solid );
	}
}

} // end of namespace picture_h
