
/*
 *  Real - Implements a real number as either a floating point or as a ratio of arbitrary
 *         sized integers
 *  Format - Stores information about how to output a number
 *  Copyright (c) 2003-2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *  See real.h
 */

/*
News
----

v4  2006-07-17 - 2006-07-30  &  2006-10-05
--

	Replaced all throw_... calls with THROW_ERROR
	Replaced all std::string usage with utf8_string
	Fixed and output error on amd64 (number containing potensis would display incorrect)

v3  2006-05-07 - 2006-05-21
--

	Some small changes in operator==, operator<=, operator>=
	Replaced constructors using float, double and long double with a constructor using floatx
	Fixed a bug in operator< and operator> which always converted ratio/integers to floatx before the comparing
	  which lead to that only the most significant numbers where compared. E.g. (2^300-1) > (2^300-10) became false
	Small clean up in floatx_to_string & Real::append_to_string

v2  2006-04-03
--

	Changed so that Real::top points to a Integer[2] and bottom to top[1], instead of them having one each
	  ( this got ump ~20% faster when dealing with small integers (integers below 2^32) )

v1  2005-09-25 - 2005-10-22
--

	Extracted class Real from the math2.* system

*/

#include "real.h"
#include "keyfile.h"


// Format-object start
void Format::make_format_legal(void)
{
	if(max_decimals > 50)
		max_decimals = 50;

	if(max_decimals != 0)
	{
		floatx newlow = powx( 10.0, -floatx(max_decimals) );

		mixed_lowlimit = (newlow > mixed_lowlimit) ? newlow : mixed_lowlimit;
	}
	else
		mixed_lowlimit = 1.0;

	if(mixed_lowlimit > mixed_hilimit)
		mixed_hilimit = mixed_lowlimit;
}

Format::Format()
{
	mixed_hilimit = 1e9;
	mixed_lowlimit = 1e-6;

	type = FormatType_Mixed;
	max_decimals = 5;
	show_trailingzeros = false;
	fraction_type = FormatFractionType_Integer_plus_fraction;
	make_format_legal();
}

void Format::set_type(FormatType newtype)
{
	if( newtype == FormatType_Normal  ||  newtype == FormatType_Mixed  ||  newtype == FormatType_Science )
		type = newtype;
}

void Format::set_fraction_type(FormatFractionType newtype)
{
	if( newtype == FormatFractionType_Fractions  ||  newtype == FormatFractionType_Integer_plus_fraction  ||
	    newtype == FormatFractionType_Decimals )
		fraction_type = newtype;
}

void Format::set_max_decimals(uint32 n)
{
	max_decimals = n;
	make_format_legal();
}

void Format::set_mixed_limits(floatx lowlimit, floatx hilimit)
{
	if(lowlimit < hilimit)
	{
		mixed_lowlimit = lowlimit;
		mixed_hilimit = hilimit;
		make_format_legal();
	}
}
// format-object end
void floatx_to_string(floatx var, utf8_string &str, const Format &fmt) throw(error_obj)
{
	int32 power = 0, extra_precision=0;
	floatx tmpvar;

	tmpvar = var;
	if( (fmt.get_type() == FormatType_Science) |
	   ( (fmt.get_type() == FormatType_Mixed) & ((tmpvar >= fmt.get_mixed_hilimit()) | (tmpvar <= -fmt.get_mixed_hilimit()) | ((tmpvar < fmt.get_mixed_lowlimit()) & (tmpvar > -fmt.get_mixed_lowlimit()) & (tmpvar != 0.0))) )
	  )
	{
		if( var >= 10.0  ||  var <= -10.0 )
		{
			for( ; var >= 10.0 ||  var <= -10.0; power++, var/=10.0 ) // calculating the 'power'
			{
				if( power > 30000 ) // a long double can only hold a power of about 4931, thus this is inf
				{
					if(var > 0.0)
						str.append( "inf" );
					else
						str.append( "-inf" );

					return;
				}
			}
		}
		else if( (var < 1.0) & (var > -1.0) & (var != 0.0) )
		{
			for( ; (var < 1.0) & (var > -1.0); power--, var*=10.0 ) // calculating the 'power'
				;
		}
	}
	else if( var < 1.0  &&  var > -1.0  &&  var != 0.0 )
	{
		for( var*=10.0; var<1.0  &&  var>-1.0; extra_precision++, var*=10.0 ) // calculating the 'power'
			;
		var = tmpvar;
	}

	{
		char out_field[100], syntax[20];
		signed int old_size = str.get_length()-1;
			
		uint64_to_char( uint64(fmt.get_max_decimals() + extra_precision), out_field );
		#ifdef USE_LONG_DOUBLE
			snprintf( syntax, 20, "%%.%sLf", out_field );
			snprintf( out_field, 100, syntax, var );
		#else
			snprintf( syntax, 20, "%%.%sf", out_field );
			snprintf( out_field, 100, syntax, double(var) );
		#endif

		str.append( out_field );

		if( fmt.get_max_decimals() != 0  &&  fmt.show_trailingzeros == false )
		{
			for( signed int i=str.get_length()-1; i>=old_size; i-- ) // removes trailing zeros
			{
				if( str.test_character( i, "0" ) )
					str.crop( i );
				else if( str.test_character( i, "." ) )
				{
					str.crop( i );
					break;
				}
				else
					break;
			}
		}


		if( (fmt.get_type() == FormatType_Science) |
		   ( (fmt.get_type() == FormatType_Mixed) & ((tmpvar >= fmt.get_mixed_hilimit()) | (tmpvar <= -fmt.get_mixed_hilimit()) | ((tmpvar < fmt.get_mixed_lowlimit()) & (tmpvar > -fmt.get_mixed_lowlimit()) & (tmpvar != 0.0))) )
		  )
		{
			char tmp[10];
			if( power >= 0 )
			{
				uint64_to_char( uint64(power), tmp );
				snprintf( out_field, 100, "*10^%s", tmp );
			}
			else
			{
				uint64_to_char( uint64(-power), tmp );
				snprintf( out_field, 100, "*10^-%s", tmp );
			}
			str.append( out_field );
		}
	}
}

// class Real start

void Real::append_to_string( utf8_string &str, const Format &fmt ) const throw(error_obj)
{
	try
	{
		if( exact )
		{
			if( fmt.get_fraction_type() == FormatFractionType_Decimals )
				floatx_to_string( floatx(*top)/floatx(*bottom), str, fmt);
			else
			{
				if((top->compare_abs_value(*bottom) ) != 0  &&  fmt.get_fraction_type() == FormatFractionType_Integer_plus_fraction)
				{
					Integer tmp_int, tmp_part;
					bool abs_over_one;

					top->divide( *bottom, &tmp_int, &tmp_part );
					if( abs_over_one = (top->compare_abs_value( *bottom ) >= 0)  ||  *top == 0 )
						tmp_int.append_to_string( str );

					if(*bottom != 1)
					{
						if( *top > 0 )
						{
							if( abs_over_one )
								str.append( " + " );
							tmp_part.append_to_string( str );
						}
						else
						{
							if( abs_over_one )
							{
								str.append( " - " );
								tmp_part.negate();
								tmp_part.append_to_string( str );
							}
							else
								tmp_part.append_to_string( str );
						}
						str.append( "/" );
						bottom->append_to_string( str );
					}
				}
				else
				{
					top->append_to_string( str );
					if(*bottom != 1  &&  *top != 0)
					{
						str.append( "/" );
						bottom->append_to_string( str );
					}
				}
			}
		}
		else
			floatx_to_string( value, str, fmt );
	}
	catch(error_obj error) {  throw error;  }
	catch(...)             {  THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );  }
}


Real::Real(int32 new_value) throw(error_obj)
{
	Integer *ptr = 0;

	try
	{
		ptr = new Integer [2];

		ptr[0] = new_value;
		ptr[1] = 1;
	}
	catch(...)
	{
		delete [] ptr;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
	exact = true;
	top = ptr;
	bottom = &ptr[1];
}

Real::Real(int64 new_value) throw(error_obj)
{
	Integer *ptr = 0;

	try
	{
		ptr = new Integer [2];

		ptr[0] = new_value;
		ptr[1] = 1;
	}
	catch(...)
	{
		delete [] ptr;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
	exact = true;
	top = ptr;
	bottom = &ptr[1];
}

Real::Real(const Integer &new_value) throw(error_obj)
{
	Integer *ptr = 0;

	try
	{
		ptr = new Integer [2];

		ptr[0] = new_value;
		ptr[1] = 1;
	}
	catch(...)
	{
		delete [] ptr;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
	exact = true;
	top = ptr;
	bottom = &ptr[1];
}

Real::Real(const Integer &new_top, const Integer &new_bottom) throw(error_obj)
{
	Integer *ptr = 0;

	try
	{
		ptr = new Integer [2];

		ptr[0] = new_top;
		ptr[1] = new_bottom;
	}
	catch(...)
	{
		delete [] ptr;
		THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
	}
	exact = true;
	top = ptr;
	bottom = &ptr[1];

	legalice();
}

Real::Real(const Real &val) throw(error_obj)
{
	if( val.exact )
	{
		Integer *ptr = 0;

		try
		{
			ptr = new Integer [2];
		}
		catch(...)
		{
			delete [] ptr;
			THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
		}

		exact = true;
		top = ptr;
		bottom = &ptr[1];

		*top = *val.top;
		*bottom = *val.bottom;
	}
	else
	{
		exact = false;
		value = val.value;
	}
}

Real::~Real()
{
	if( exact )
		delete [] top;
}

void Real::legalice(void)
{
	if( exact )
	{
		if( *bottom < 0 )
		{
			top->negate();
			bottom->negate();
		}
		else if( *bottom == 0 )
		{
			*bottom = 1;
			*top = 0;
			return;
		}

		if( *bottom == 1 )
			return;

		if( *top == 0 )
			*bottom = 1;
		else
		{
			Integer div = gcd( *top, *bottom ); // lets simplify the fraction
			if( div != 0 )
			{
				Integer tmp;

				top->divide( div, &tmp, 0 );
				top->swap( tmp );
				bottom->divide( div, &tmp, 0 );
				bottom->swap( tmp );
			}
		}
	}
}

Real Real::operator+(const Real &val) const throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
		return Real( floatx(*this) + floatx(val) );
	else // exact fraction calculus
	{
		if( *bottom == *val.bottom )
		{
			Real res( *top + *val.top, *bottom );
			res.legalice();
			return res;
		}
		else
		{
			Real res( *top * *val.bottom + *val.top * *bottom, *bottom * *val.bottom );
			res.legalice();
			return res;
		}
	}
}
void Real::add(const Real &val) throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
	{
		if( exact )
		{
			floatx tmp;
			tmp = floatx(*this);
			delete [] top;
			exact = false;
			value = tmp;
		}
		value += floatx(val);
	}
	else // exact fraction calculus
	{
		if( *bottom == *val.bottom )
			top->add( *val.top );
		else
		{
			Integer tmp;
			top->mul( *val.bottom, &tmp );
			val.top->mul( *bottom, top );
			top->add( tmp );

			bottom->mul( *val.bottom, &tmp );
			bottom->swap( tmp );
		}
		legalice();
	}
}

Real Real::operator-(const Real &val) const throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
		return Real( floatx(*this) - floatx(val) );
	else // exact fraction calculus
	{
		if( *bottom == *val.bottom )
		{
			Real res( *top - *val.top, *bottom );
			res.legalice();
			return res;
		}
		else
		{
			Integer a,b;
			top->mul( *val.bottom, &a );
			val.top->mul( *bottom, &b );
			a -= b;
			bottom->mul( *val.bottom, &b );
			{
				Real res( a, b );
				res.legalice();
				return res;
			}
		}
	}
}
void Real::sub(const Real &val) throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
	{
		if( exact )
		{
			floatx tmp;
			tmp = floatx(*this);
			delete [] top;
			exact = false;
			value = tmp;
		}
		value -= floatx(val);
	}
	else // exact fraction calculus
	{
		if( *bottom == *val.bottom )
			top->sub( *val.top );
		else
		{
			Integer tmp;
			top->mul( *val.bottom, &tmp );
			val.top->mul( *bottom, top );
			top->negate();
			top->add( tmp );

			bottom->mul( *val.bottom, &tmp );
			bottom->swap( tmp );
		}
		legalice();
	}
}

Real operator-(const Real &val) throw(error_obj)
{
	if( val.exact )
	{
		if( *val.top == 0 )
			return val;
		return Real( -*val.top, *val.bottom );
	}
	else
	{
		if( val.value == 0.0 )
			return val;
		return Real( -val.value );
	}
}

Real Real::operator*(const Real &val) const throw(error_obj)
{
	if(exact == false  ||  val.exact == false) // floating point calculus
		return Real( floatx(*this) * floatx(val) );
	else // exact fraction calculus
	{
		Integer a,b;
		top->mul( *val.top, &a );
		bottom->mul( *val.bottom, &b );
		{
			Real res( a, b );
			res.legalice();

			return res;
		}
	}
}
void Real::mul(const Real &val) throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
	{
		if( exact )
		{
			floatx tmp;
			tmp = floatx(*this);
			delete [] top;
			exact = false;
			value = tmp;
		}
		value *= floatx(val);
	}
	else // exact fraction calculus
	{
		Integer tmp;
		top->mul( *val.top, &tmp );
		top->swap( tmp );
		bottom->mul( *val.bottom, &tmp );
		bottom->swap( tmp );
		legalice();
	}
}

Real Real::operator/(const Real &val) const throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating pointcalculus
		return Real( floatx(*this) / floatx(val) );
	else // exact fraction calculus
	{
		Real res( *top * *val.bottom, *bottom * *val.top );
		res.legalice();
		return res;
	}
}
void Real::div(const Real &val) throw(error_obj)
{
	if( exact == false  ||  val.exact == false ) // floating point calculus
	{
		if( exact )
		{
			floatx tmp;
			tmp = floatx(*this);
			delete [] top;
			exact = false;
			value = tmp;
		}
		value /= floatx(val);
	}
	else // exact fraction calculus
	{
		Integer tmp;
		top->mul( *val.bottom, &tmp );
		top->swap( tmp );
		bottom->mul( *val.top, &tmp );
		bottom->swap( tmp );
		legalice();
	}
}

Real & Real::operator=(const Real &val) throw(error_obj)
{
	if( val.exact )
	{
		if( !exact )
		{
			Integer *ptr = 0;

			try
			{
				ptr = new Integer [2];
			}
			catch(...)
			{
				delete [] ptr;
				THROW_ERROR( ErrorType_Memory, _("Couldn't get memory.") );
			}

			exact = true;
			top = ptr;
			bottom = &ptr[1];
		}

		*top = *val.top;
		*bottom = *val.bottom;
	}
	else
	{
		if( exact )
		{
			delete [] top;
			exact = false;
		}
		value = val.value;
	}
	return *this;
}

bool Real::operator==(const Real &val) const
{
	if( exact != val.exact )
		return ( floatx(*this) == floatx(val) );
	else
	{
		if( exact )
			return ( *top == *val.top  &&  *bottom == *val.bottom );
		else
			return ( value == val.value );
	}
}

bool Real::operator<(const Real &val) const
{
	if( exact != val.exact )
		return ( floatx(*this) < floatx(val) );
	else
	{
		if( exact )
		{
			if( *bottom == *val.bottom )
				return *top < *val.top;
			else
			{
				Integer tmp, tmp2;
				top->mul( *val.bottom, &tmp ); // we must have the same divider of the ratios
				val.top->mul( *bottom, &tmp2 );

				return tmp < tmp2;
			}
		}
		else
			return ( value < val.value );
	}
}

bool Real::operator>(const Real &val) const
{
	if( exact != val.exact )
		return ( floatx(*this) > floatx(val) );
	else
	{
		if( exact )
		{
			if( *bottom == *val.bottom )
				return *top > *val.top;
			else
			{
				Integer tmp, tmp2;
				top->mul( *val.bottom, &tmp ); // we must have the same divider of the ratios
				val.top->mul( *bottom, &tmp2 );

				return tmp > tmp2;
			}
		}
		else
			return ( value > val.value );
	}
}


Real ipart(const Real &val) throw(error_obj)
{
	if( val.exact )
		return Real( *val.top / *val.bottom );
	else
		return Real( truncx(val.value) );
}

Real fpart(const Real &val) throw(error_obj)
{
	if( val.exact )
		return Real( *val.top % *val.bottom, *val.bottom );
	else
		return Real( val.value - truncx(val.value) );
}

Real sign(const Real &val) throw(error_obj)
{
	if(val.exact)
	{
		int32 res;

		if( *val.top == 0 )
			res = 0;
		else
			res = (*val.top < 0) ? -1 : 1;
		return Real( res );
	}
	else
	{
		floatx res;

		if( val.value == 0.0 )
			res = 0.0;
		else
			res = (val.value < 0.0) ? -1.0 : 1.0;
		return Real( res );
	}
}

// converts a Real to a fraction
Real frac(const Real &val, int32 highest_bottom_value) throw(error_obj)
{
	if( highest_bottom_value < 1 )
		THROW_ERROR( ErrorType_Domain, _("Domain error: Value out of range.") );

	if(val.exact)
	{
		if( *val.bottom <= highest_bottom_value )
			return val;
		else
			return frac( Real( floatx(val) ), highest_bottom_value );
	}
	else
	{
		Real res;
		floatx best_fault, test_fault, tmp, value;
		bool negative;
		int32 top, bottom;

		if( val.value < 0.0 )
		{
			negative = true;
			value = -val.value;
		}
		else
		{
			negative = false;
			value = val.value;
		}
		value = value - truncx(value);
		res = 0;

		best_fault = value;

		if(best_fault != 0.0)
		{
			floatx fbottom;

			if( (bottom = highest_bottom_value / 2) == 0 )
				bottom = 1;

			top = int32( value * floatx(bottom) );

			fbottom = floatx(bottom);

			test_fault = value - (floatx(top)/fbottom);
			do
			{
				if(test_fault > 0.0 )
					top++;
				else
				{
					if( bottom == highest_bottom_value )
						break;
					bottom++;
					fbottom = floatx(bottom);
					top--;
				}

				test_fault = value - (floatx(top)/fbottom);
				tmp = (test_fault < 0.0) ? -test_fault : test_fault;

				if(tmp < best_fault)
				{
					best_fault = tmp;
					*res.top = top;
					*res.bottom = bottom;
				}
			}while(test_fault != 0.0);
		}

		{
			Integer tmp1, tmp2;
			if( negative )
				tmp1 = -val.value;
			else
				tmp1 = val.value;
			tmp1.mul( *res.bottom, &tmp2 );
			res.top->add( tmp2 );
			res.legalice();
		}

		if( negative )
			res.top->negate();

		return res;
	}
}

Real power(const Real &val, const Integer &nr) throw(error_obj)
{
	Real res(0);

	if( !val.isExact() ) // this error should never accour...that's why it's not translatable
		THROW_ERROR( ErrorType_Domain, _("Domain error: Real power(val, nr), val must be a ratio.") );

	if( val.isInteger() )
	{
		if( nr.isNegative() )
		{ // n^p   p<0
			Integer tmp, tnr;

			tnr = nr;
			tnr.negate();

			val.top->power( tnr, &tmp );

			*res.top = 1;
			*res.bottom = tmp;
		}
		else
		{ // n^p   p>=0
			Integer tmp;

			val.top->power( nr, &tmp );
			*res.top = tmp;
			*res.bottom = 1;
		}
	}
	else
	{
		if( nr.isNegative() )
		{ // (n/m)^p   p<0
			Integer tmp, tnr;

			tnr = nr;
			tnr.negate();

			val.top->power( tnr, &tmp );
			*res.bottom = tmp;

			val.bottom->power( tnr, &tmp );
			*res.top = tmp;
		}
		else
		{ // (n/m)^p   p>=0
			Integer tmp;

			val.top->power( nr, &tmp );
			*res.top = tmp;

			val.bottom->power( nr, &tmp );
			*res.bottom = tmp;
		}
	}
	return res;
}
