
/*
 *  Real - Implements a real number as either a floating point or as a ratio of arbitrary
 *         sized integers
 *  Format - Stores information about how to output a Real
 *  Copyright (c) 2005 - 2006 by Mattias Hultgren <mattias_hultgren@tele2.se>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */


#ifndef REAL_H_
#define REAL_H_

#include "vartypes.h"
#include "integer.h"
#include "utf8_string.h"


#define REAL_H_VERSION "v4"
#define REAL_H_DATE    "2005-09 - 2006-10"




enum FormatType{ FormatType_Normal, FormatType_Mixed, FormatType_Science };
enum FormatFractionType{ FormatFractionType_Fractions, FormatFractionType_Integer_plus_fraction,
                         FormatFractionType_Decimals };
class Format
{
private:
	floatx mixed_hilimit;
	floatx mixed_lowlimit;

	FormatType type;
	FormatFractionType fraction_type;
	uint32 max_decimals;

	void make_format_legal(void);

public:
	bool show_trailingzeros;

	Format();
	void set_type(FormatType newtype);
	inline FormatType get_type(void) const { return type; }

	void set_fraction_type(FormatFractionType newtype);
	inline FormatFractionType get_fraction_type(void) const { return fraction_type; }

	void set_max_decimals(uint32 n);
	inline uint32 get_max_decimals(void) const { return max_decimals; }


	// if lowlimit is higher than hilimit nothing will happen
	void set_mixed_limits(floatx lowlimit, floatx hilimit);
	inline floatx get_mixed_lowlimit(void) const { return mixed_lowlimit; }
	inline floatx get_mixed_hilimit(void) const { return mixed_hilimit; }
};

// if an exception is throwed the result in str may contain partly finished text
void floatx_to_string(floatx var, utf8_string &str, const Format &fmt) throw(error_obj);



class Real
{
private:
	bool exact;
	union{
		floatx value;
		struct{
			Integer *top, *bottom;
		};
	};

	void legalice(void);

public:
	Real() { exact = false; value = 0.0; }
	Real(const floatx &new_value) { exact = false; value = floatx(new_value); }
	Real(int32 new_value) throw(error_obj);
	Real(int64 new_value) throw(error_obj);
	Real(const Integer &new_value) throw(error_obj);
	Real(const Integer &new_top, const Integer &new_bottom) throw(error_obj);
	Real(const Real &val) throw(error_obj);
	~Real();


	void add( const Real &val ) throw(error_obj);
	void sub( const Real &val ) throw(error_obj);
	void mul( const Real &val ) throw(error_obj);
	void div( const Real &val ) throw(error_obj);

	inline bool isInteger(void) const { if( exact ) return ( *bottom == 1 ); else return false; }
	inline bool isExact(void) const { return exact; }
	inline void get_top( Integer *ptr ) const { if(exact) *ptr = *top; }
	inline void get_bottom( Integer *ptr ) const { if(exact) *ptr = *bottom; }
	inline operator floatx(void) const { if(exact) return floatx(*top)/floatx(*bottom); else return value; }

	inline Real & operator++(void) throw(error_obj) { if(exact) top->add( *bottom ); else value++; return *this; }
	inline Real & operator++(int)  throw(error_obj) { if(exact) top->add( *bottom ); else value++; return *this; }
	inline Real & operator--(void) throw(error_obj) { if(exact) top->sub( *bottom ); else value--; return *this; }
	inline Real & operator--(int)  throw(error_obj) { if(exact) top->sub( *bottom ); else value--; return *this; }

	Real operator+(const Real &val) const throw(error_obj);
	Real operator-(const Real &val) const throw(error_obj);
	Real operator*(const Real &val) const throw(error_obj);
	Real operator/(const Real &val) const throw(error_obj);

	Real & operator=(const Real &val) throw(error_obj);
	inline Real & operator+=(const Real &val) throw(error_obj) { this->add( val ); return *this; }
	Real & operator-=(const Real &val) throw(error_obj) { this->sub( val ); return *this; }
	Real & operator*=(const Real &val) throw(error_obj) { this->mul( val ); return *this; }
	inline Real & operator*=(const int64 &val) throw(error_obj) { if(exact){ *top*=val; legalice(); }else value*=val; return *this;}
	inline Real & operator*=(const float &val) throw(error_obj) { if(exact){ value=floatx(*this)*val; exact=false; }else value*=val; return *this;}
	inline Real & operator*=(const double &val) throw(error_obj) { if(exact){ value=floatx(*this)*val; exact=false; }else value*=val; return *this;}
	inline Real & operator*=(const long double &val) throw(error_obj) { if(exact){ value=floatx(*this)*val; exact=false; }else value*=val; return *this;}
	Real & operator/=(const Real &val) throw(error_obj) { this->div( val ); return *this; }


	bool operator==(const Real &val) const;
	inline bool operator!=(const Real &val) const { return !(*this == val); };

	bool operator<(const Real &val) const;
	bool operator>(const Real &val) const;
	inline bool operator<=(const Real &val) const { return !(*this > val); }
	inline bool operator>=(const Real &val) const { return !(*this < val); }

	void append_to_string( utf8_string &str, const Format &fmt ) const throw(error_obj);

	friend Real operator-(const Real &val) throw(error_obj);
	friend Real ipart(const Real &val) throw(error_obj);
	friend Real fpart(const Real &val) throw(error_obj);
	// converts a Real to a fraction
	friend Real frac(const Real &val, int32 highest_bottom_value = 1000) throw(error_obj);
	friend Real sign(const Real &val) throw(error_obj);

	// val.isExact() must return true
	friend Real power(const Real &val, const Integer &nr) throw(error_obj);
};






#endif // REAL_H_
